#include "game/landscape.h"
#include "application.h"
// note: also adjust this in the shader
// TODO:
//
// What we want is, for gameplay, two maps spanning the entire world
// 1. height
// 2. type (water, grass, ...)
//
// For rendering, we combine this into one RGBA map, storing the height in
// the alpha component:
//
// RGBA map: height+color+light
//
// Now, any change to height has to be synchronized back to the RGBA map, in
// the alpha component.
//
// And that change also has to update the light (from the height map).
//
// A vertex shader actually could perform the lighting, with no need to do it
// on the CPU and write into the map.
//
// But otherwise, a shader simply using the texture for the height for rgb works.
// All we have to ensure is to whenever height changes, update the height as well
// as the color+light and synchronize the affected areas of each back to the GPU.
//
// All lighting is done by vertex colors. Each vertex gets the color by
// multiplying the rgb triplet stored as vertex attribute with another rgb
// value read from a texture.
//
// TODO: if there is one color per vertex either way, why use vertex attributes
//       as well as the texture?
//
// TODO: in fact at this point, we can get rid of the whole chunk loading
//       alltogether. Just have a fixed grid on the GPU, reading height and
//       color information from the texture.
//
// TODO: We limit the size of the world to the big texture - eventually we
//       may want the chunk loading for that, keeping several textures instead
//       of a single one for the entire world.
//
// The texture is twice as wide and high as the game's heightmap.
// It is arranged like this:
//
// V . V . V . V .
// . x . x . x . x
// V . V . V . V .
// . x . x . x . x
//
// V are the actual height values stored in the tilemap.
// x are averaged vertices created from their 4 neighbors in fill_chunk below.
// . are extra texels not directly used.
//
// The reason is that the average 'x' positions allow for smoother
// interpolation.
//
// tile-position 0/0 is in the center of the texture.
#if defined(ANDROID) || defined(MACOS) || defined(SIMPLE_SHADERS)
static char const * vertex_glsl = "\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "#define LIGHTMAP_RESOLUTION 1\n"
    "attribute vec4 xyz;\n"
    "attribute vec4 rgba;\n"
    "uniform mat4 modelview_matrix, projection_matrix;\n"
    "uniform sampler2D tex;\n"
    "uniform float lightscale;\n"
    "uniform float max_height;\n"
    "varying vec4 color;\n"
    "void main(void) {\n"
    "    gl_Position = projection_matrix * modelview_matrix * xyz;\n"
    "    color = rgba; // * vec3(texcol);\n"
    "}\n"
    "";
static char const * fragment_glsl = "\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "varying vec4 color;\n"
    "void main(void) {\n"
    "    if (color.a < 0.01) discard;\n"
    "    gl_FragColor = color;\n"
    "}\n"
    "";
#else
static char const * vertex_glsl = "\n"
    "#version 300 es // for fragment shader\n"
    "//#version 120\n"
    "#define LIGHTMAP_RESOLUTION 1\n"
    "in vec4 xyz;\n"
    "in vec3 rgb;\n"
    "in float wireframe;\n"
    "uniform mat4 modelview_matrix, projection_matrix;\n"
    "uniform sampler2D tex;\n"
    "uniform float lightscale;\n"
    "uniform float max_height;\n"
    "out vec3 color;\n"
    "out vec3 barycentric;\n"
    "uniform bool use_wireframe;\n"
    "void main(void) {\n"
    "    //vec2 uv = vec2(xyz) * lightscale;\n"
    "    //uv.x += lightscale / 2 / LIGHTMAP_RESOLUTION;\n"
    "    //uv.y -= lightscale / 2 / LIGHTMAP_RESOLUTION;\n"
    "    //vec4 texcol = texture2D(tex, uv);\n"
    "    // TODO: just a test to demonstrate reading a height offset from a texture\n"
    "    //vec4 xyz2 = xyz; //vec4(vec2(xyz), texcol.w * max_height, 1);\n"
    "    gl_Position = projection_matrix * modelview_matrix * xyz;\n"
    "    color = rgb; // * vec3(texcol);\n"
    "    if (use_wireframe) {\n"
    "        vec3 a[3] = vec3[](vec3(1, 0, 0), vec3(0, 1, 0), vec3(0, 0, 1));\n"
    "        barycentric = a[int(wireframe)];\n"
    "    }\n"
    "    else {\n"
    "        barycentric = vec3(0, 0, 0);\n"
    "    }\n"
    "}\n"
    "";
static char const * fragment_glsl = "\n"
    "#version 300 es // for fwidth\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "in vec3 color;\n"
    "out vec4 out_color;\n"
    "in vec3 barycentric;\n"
    "uniform bool use_wireframe;\n"
    "float edgeFactor() {\n"
    "    vec3 d = fwidth(barycentric);\n"
    "    vec3 a3 = smoothstep(vec3(0.0), d*1.5, barycentric);\n"
    "    return min(min(a3.x, a3.y), a3.z);\n"
    "}\n"
    "\n"
    "void main(void) {\n"
    "    if (use_wireframe) {\n"
    "        float f = 1.0 - edgeFactor();\n"
    "        out_color = vec4(color.xyz * f, f);\n"
    "    }\n"
    "    else {\n"
    "        out_color = vec4(color, 1.0);\n"
    "    }\n"
    "}\n"
    "";
#endif
static void color_water(float c, float * r, float * g, float * b);
static void color_grass(float c, float * r, float * g, float * b);
static void color_mountain(float c, float * r, float * g, float * b);
static void color_snow(float c, float * r, float * g, float * b);
static void light_tops(Landscape * self);
static float _normalized_noise_to_height(Landscape * self, float v);
static float _height_to_normalized_noise(Landscape * self, float h);
static LandColor _color_for_height(Landscape * self, float h);
static Tile* _get_tile(Landscape * self, int x, int y);
static float get_height(Landscape * self, int x, int y);
static bool get_color(Landscape * self, int x, int y, GLfloat * r, GLfloat * g, GLfloat * b, GLfloat * a);
static void fill_chunk(Landscape * self, Chunk * chunk);
static LandVector _get_point(Landscape * self, Camera * cam, int x, int y);
static void color_water(float c, float * r, float * g, float * b) {
    * r = 0.3 + 0.25 * c;
    * g = 0.5 + 0.5 * c;
    * b = 1;
}
static void color_grass(float c, float * r, float * g, float * b) {
    * r = 0.55 + 0.25 * c;
    * g = 1;
    * b = 0.3 + 0.1 * c;
}
static void color_mountain(float c, float * r, float * g, float * b) {
    * r = 0.8 + 0.2 * c;
    * g = 0.2 + 0.4 * c;
    * b = 0.1 + 0.2 * c;
}
static void color_snow(float c, float * r, float * g, float * b) {
    * r = 0.6 + 0.4 * c;
    * g = 0.6 + 0.4 * c;
    * b = 0.8 + 0.2 * c;
}
static void light_tops(Landscape * self) {
    float light_top = self->max_height - 2;
    int w = self->w;
    for (int y = 0; y < w; y += 1) {
        for (int x = 0; x < w; x += 1) {
            Tile * t = self->heightmap + x + y * w;
            if (t->height > light_top) {
                float f = (t->height - light_top) / 2;
                t->dark = t->dark + (f * f * f * f * 0.5) * 255;
                if (t->dark > 255) {
                    t->dark = 255;
                }
            }
        }
    }
}
Landscape* landscape_create(int logsize, int seed) {
    return landscape_create_chunks(logsize, 64, seed);
}
Landscape* landscape_create_chunks(int logsize, int chunks, int seed) {
    Landscape * l = landscape_create_base();
    l->chunks_count = chunks;
    l->w = 1 << logsize;
    landscape_create_map(l, seed);
    return l;
}
Landscape* landscape_create_base(void) {
    Landscape * self;
    land_alloc(self);
    self->adjustments = land_array_new();
    self->chunks_count = 64;
    self->want_light_tops = 1;
    self->want_rivers = 1;
    self->max_height = 20;
    self->snow_top = 10;
    self->plateau_levels = 0;
    self->water_levels = 100;
    self->chunk_s = 32;
    self->vertices_per_tile = 12;
    self->stride = 8 /* x y z r g b a wireframe */;
    self->max_visible_chunks_s = 8;
    self->visible_chunks_w = 8;
    self->visible_chunks_h = 8;
    self->river_color = land_color_rgba(0.3, 0.5, 1, 1);
    self->fog_color = land_color_rgba(0.5, 0.5, 0.5, 1);
    self->color_for_height = _color_for_height;
    self->blur = 8;
    return self;
}
void landscape_create_map(Landscape * self, int seed) {
    landscape_create_map_heightmap(self, seed);
    landscape_create_colors(self);
}
void landscape_create_map_flat(Landscape * self, float h) {
    int w = self->w;
    self->heightmap = land_calloc(w * w * sizeof (* self->heightmap));
    for (int y = 0; y < w; y += 1) {
        for (int x = 0; x < w; x += 1) {
            Tile * tile = self->heightmap + y * w + x;
            tile->height = h;
        }
    }
    isles_update_light(self);
}
static float _normalized_noise_to_height(Landscape * self, float v) {
    print("%f", v);
    float tm = self->water_levels + 255 + self->plateau_levels;
    float h = v * tm /* -tm..tm */;
    h -= self->water_levels;
    if (h < 0) {
        h = 0;
    }
    if (h > 255) {
        h = 255;
    }
    return self->max_height * h / 255;
}
static float _height_to_normalized_noise(Landscape * self, float h) {
    float tm = self->water_levels + 255 + self->plateau_levels;
    float v = h * 255 / self->max_height;
    v += self->water_levels;
    v /= tm;
    return v;
}
void landscape_create_map_heightmap(Landscape * self, int seed) {
    int w = self->w;
    self->heightmap = land_calloc(w * w * sizeof (* self->heightmap));
    // limit is -1..+1, so we divide amplitude by half the maximum value
    float t = (self->water_levels + 255 + self->plateau_levels) / 2;
    LandNoise * noise = land_noise_new(LandNoisePlasma, seed);
    land_noise_set_size(noise, w, w);
    land_noise_set_levels(noise, 7);
    land_noise_set_amplitude(noise, log2(8 * w / 2 / t));
    LandNoise * blur = land_noise_new(LandNoiseValue, 0);
    land_noise_set_random(blur, noise->seed);
    land_noise_set_blur(blur, noise, self->blur);
    land_noise_prepare(blur);
    landscape_process_noise(self, blur);
}
void landscape_process_noise(Landscape * self, LandNoise * blur) {
    int w = self->w;
    float water_height = _height_to_normalized_noise(self, 0);
    float snow_height = _height_to_normalized_noise(self, 11);
    for (int y = 0; y < w; y += 1) {
        for (int x = 0; x < w; x += 1) {
            Tile * tile = self->heightmap + y * w + x;
            float v = land_noise_at(blur, x, y);
            tile->height = _normalized_noise_to_height(self, v);
        }
    }
    if (self->want_rivers) {
        LandNoise * r = rivers(blur, water_height, snow_height);
        for (int y = 0; y < w; y += 1) {
            for (int x = 0; x < w; x += 1) {
                Tile * tile = self->heightmap + y * w + x;
                tile->river = land_noise_at(r, x, y);
            }
        }
        land_noise_destroy(r);
    }
    landscape_calculate_light(self);
}
void landscape_calculate_light(Landscape * self) {
    isles_update_light(self);
    if (self->want_light_tops) {
        light_tops(self);
    }
}
#define P(x, y) \
    rgba + (((y) & (w - 1)) * w + ((x) & (w - 1))) * 4;
static LandColor _color_for_height(Landscape * self, float h) {
    float const snow_height = 9.75;
    float const snow_height2 = self->snow_top;
    float f;
    float r1, g1, b1, r2, g2, b2;
    if (h < 1) {
        f = h;
        color_water(f, & r1, & g1, & b1);
        color_grass(1 - f, & r2, & g2, & b2);
    }
    else if (h < 5) {
        f = 0;
        color_grass(0, & r1, & g1, & b1);
        color_grass(0, & r2, & g2, & b2);
    }
    else if (h < 6) {
        f = h - 5;
        color_grass(0, & r1, & g1, & b1);
        color_mountain(1, & r2, & g2, & b2);
    }
    else if (h < snow_height) {
        f = 0;
        color_mountain(1, & r1, & g1, & b1);
        color_mountain(1, & r2, & g2, & b2);
    }
    else if (h < snow_height2) {
        f = (h - snow_height) / (snow_height2 - snow_height);
        color_mountain(1, & r1, & g1, & b1);
        color_snow(1, & r2, & g2, & b2);
    }
    else {
        f = 1;
        color_mountain(1, & r1, & g1, & b1);
        color_snow(1, & r2, & g2, & b2);
    }
    float r = r2 * f + r1 * (1 - f);
    float g = g2 * f + g1 * (1 - f);
    float b = b2 * f + b1 * (1 - f);
    float a = 1 /* h / self.max_height */;
    return land_color_rgba(r, g, b, a);
}
void landscape_create_colors(Landscape * self) {
    int s = LIGHTMAP_RESOLUTION;
    int w = self->w * s;
    unsigned char * rgba = land_malloc(w * w * 4);
    for (int y = 0; y < w; y += s) {
        // the rgba data will have the origin in the lower left corner,
        // so y is flipped
        for (int x = 0; x < w; x += s) {
            float hx = x / (float) s, hy = - y / (float) s;
            float h = landscape_get_height_interpolated(self, hx, hy);
            unsigned char * p = rgba + (y * w + x) * 4;
            LandColor c = self->color_for_height(self, h);
            p [0] = 255 * c.r;
            p [1] = 255 * c.g;
            p [2] = 255 * c.b;
            p [3] = 255 * c.a;
        }
    }
    // lightscale is set to 1/w, so the maps x/y coordinate is directly
    // the texture coordinate. Since our texture is twice the resolution
    // that means integer map coordinates are all the even pixels.
    //
    // there also is an offset of 0.25/w which means each vertex falls
    // right into the center of one of our pixels.
    //
    // In the picture below 0, 2, 6, 8 are the four pixel at full integer
    // coordinates. The center of each has a vertex. Pixel 4 has the
    // center vertex at its center. 1, 3, 5, 7 have no effect as they
    // should be out of interpolation distance to all five vertices.
    //
    //
    //  \ | / \ | / #
    //   \|/   \|/  #
    //  __0__1__2__ #
    //   /|\   /|\  #
    //  / | \ / | \ #
    //    3  4  5   #
    //  \ | / \ | / #
    //   \|/   \|/  #
    //  __6__7__8__ #
    //   /|\   /|\  #
    //  / | \ / | \ #
    //              #
    if (s == 2) {
        for (int y = 0; y < w; y += 2) {
            for (int x = 0; x < w; x += 2) {
                unsigned char * p0 = P(x, y);
                unsigned char * p1 = P(x + 1, y);
                unsigned char * p2 = P(x + 2, y);
                unsigned char * p3 = P(x + 0, y + 1);
                unsigned char * p4 = P(x + 1, y + 1);
                unsigned char * p6 = P(x + 0, y + 2);
                unsigned char * p8 = P(x + 2, y + 2);
                for (int i = 0; i < 3; i += 1) {
                    p1 [i] = (p0 [i] + p2 [i]) / 2;
                    p3 [i] = (p0 [i] + p6 [i]) / 2;
                    unsigned char p5 = (p2 [i] + p8 [i]) / 2;
                    p4 [i] = (p3 [i] + p5) / 2;
                }
            }
        }
    }
    int deserts = 5;
    for (int i = 0; i < deserts; i += 1) {
        int cx = land_rand(0, w);
        int cy = land_rand(0, w);
        int radius = 50 * s;
        for (int yc = cy - radius; yc < cy + radius + 1; yc += 1) {
            for (int xc = cx - radius; xc < cx + radius + 1; xc += 1) {
                Tile * tile = landscape_get_tile(self, xc / (float) s, - yc / (float) s);
                int x = xc % w;
                int y = yc % w;
                if (x < 0) {
                    x += w;
                }
                if (y < 0) {
                    y += w;
                }
                unsigned char * p = rgba + (y * w + x) * 4;
                float r = p [0] / 255.0;
                float g = p [1] / 255.0;
                float b = p [2] / 255.0;
                //float a = p[3] / 255.0 * self.max_height
                float a = landscape_get_height_interpolated(self, x / (float) s, - y / (float) s);
                float dx = xc - cx;
                float dy = yc - cy;
                float d = sqrt(dx * dx + dy * dy);
                float f = (radius - d) / 10.0;
                if (f < 0) {
                    f = 0;
                }
                if (f > 1) {
                    f = 1;
                }
                if (a < 1) {
                    f *= a;
                }
                if (a > 5) {
                    float a2 = (a - 5) / 2.0;
                    if (a2 > 1) {
                        a2 = 1;
                    }
                    f *= (1 - a2);
                }
                float r2 = 0.9;
                float g2 = 0.8;
                float b2 = 0.1;
                r = r * (1 - f) + r2 * f;
                g = g * (1 - f) + g2 * f;
                b = b * (1 - f) + b2 * f;
                p [0] = 255 * r;
                p [1] = 255 * g;
                p [2] = 255 * b;
                if (f > 0.5) {
                    tile->kind = TK_DESERT;
                }
            }
        }
    }
    self->lightmap_rgba = rgba;
}
static Tile* _get_tile(Landscape * self, int x, int y) {
    int wm = self->w - 1;
    return self->heightmap + (y & wm) * self->w + (x & wm);
}
Tile* landscape_get_tile(Landscape * self, int x, int y) {
    return _get_tile(self, x, y);
}
static float get_height(Landscape * self, int x, int y) {
    Tile * tile = _get_tile(self, x, y);
    return tile->height;
}
float landscape_get_height_interpolated(Landscape * self, F px, F py) {
    int tx = floor(px);
    int ty = floor(py);
    F x_ = tx + 0.5;
    F y_ = ty + 0.5;
    F x = px - x_;
    F y = py - y_;
    F x1, y1, x2, y2;
    float h [4];
    for (int i = 0; i < 4; i += 1) {
        h [i] = get_height(self, tx + (i & 1), ty + (i >> 1));
    }
    F h3 = (h [0] + h [1] + h [2] + h [3]) / 4;
    F h2, h1;
    if (x > y) {
        // upper right half
        x2 = 0.5;
        y2 = - 0.5;
        h2 = h [1];
    }
    else {
        // lower left half
        x2 = - 0.5;
        y2 = 0.5;
        h2 = h [2];
    }
    if (x > - y) {
        // lower right half
        x1 = 0.5;
        y1 = 0.5;
        h1 = h [3];
    }
    else {
        // upper left half
        x1 = - 0.5;
        y1 = - 0.5;
        h1 = h [0];
    }
    // barycentric coordinates
    // p = v1 * g1 + v2 * g2 + v3 * g3, where g1 + g2 + g3 = 1
    // separated:
    // x = x1 * g1 + x2 * g2 + x3 * g3
    // y = y1 * g1 + y2 * g2 + y3 * g3
    // we use x3 = 0, y3 = 0
    // x = x1 * g1 + x2 * g2
    // y = y1 * g1 + y2 * g2
    // therefore
    // g1 * x1 = x - x2 * g2
    // g2 * y2 = y - y1 * g1
    F g1 = (y2 * x - x2 * y) / (y2 * x1 - x2 * y1);
    F g2 = (y1 * y - y1 * x) / (y2 * x1 - x2 * y1);
    F g3 = 1 - g1 - g2;
    return h1 * g1 + h2 * g2 + h3 * g3;
}
void landscape_get_normal_interpolated(Landscape * self, F px, F py, F * rnx, F * rny, F * rnz) {
    int tx = floor(px);
    int ty = floor(py);
    F x_ = tx + 0.5;
    F y_ = ty + 0.5;
    F x = px - x_;
    F y = py - y_;
    F x1, y1, x2, y2;
    float nx [4], ny [4], nz [4];
    float nx3 = 0, ny3 = 0, nz3 = 0;
    for (int i = 0; i < 4; i += 1) {
        Tile * t = _get_tile(self, tx + (i & 1), ty + (i >> 1));
        nx [i] = t->nx / 100.0;
        ny [i] = t->ny / 100.0;
        nz [i] = t->nz / 100.0;
        nx3 += nx [i];
        ny3 += ny [i];
        nz3 += nz [i];
    }
    int i1, i2;
    if (x > y) {
        // upper right half
        x2 = 0.5;
        y2 = - 0.5;
        i2 = 1;
    }
    else {
        // lower left half
        x2 = - 0.5;
        y2 = 0.5;
        i2 = 2;
    }
    if (x > - y) {
        // lower right half
        x1 = 0.5;
        y1 = 0.5;
        i1 = 3;
    }
    else {
        // upper left half
        x1 = - 0.5;
        y1 = - 0.5;
        i1 = 0;
    }
    F g1 = (y2 * x - x2 * y) / (y2 * x1 - x2 * y1);
    F g2 = (y1 * y - y1 * x) / (y2 * x1 - x2 * y1);
    F g3 = 1 - g1 - g2;
    float n3 = sqrt(nx3 * nx3 + ny3 * ny3 + nz3 * nz3);
    nx3 /= n3;
    ny3 /= n3;
    nz3 /= n3;
    float rx, ry, rz;
    rx = nx3 * g3 + nx [i1] * g1 + nx [i2] * g2;
    ry = ny3 * g3 + ny [i1] * g1 + ny [i2] * g2;
    rz = nz3 * g3 + nz [i1] * g1 + nz [i2] * g2;
    float r = sqrt(rx * rx + ry * ry + rz * rz);
    if (r > 0.01) {
        * rnx = rx / r;
        * rny = ry / r;
        * rnz = rz / r;
    }
    else {
        * rnx = 0;
        * rny = 0;
        * rnz = 1;
    }
}
static bool get_color(Landscape * self, int x, int y, GLfloat * r, GLfloat * g, GLfloat * b, GLfloat * a) {
    Tile * tile = _get_tile(self, x, y);
    float d = tile->dark / 255.0;
    LandColor c = self->color_for_height(self, tile->height);
    float p = tile->river;
    if (p > 0) {
        c.r *= (1 - p);
        c.g *= (1 - p);
        c.b *= (1 - p);
        c.r += p * self->river_color.r;
        c.g += p * self->river_color.g;
        c.b += p * self->river_color.b;
    }
    bool debug = app_global ()->debug_reveal_map;
    float h = debug ? 1 : tile->visible / 255.0;
    * r = c.r * d * h + self->fog_color.r * (1 - h);
    * g = c.g * d * h + self->fog_color.g * (1 - h);
    * b = c.b * d * h + self->fog_color.b * (1 - h);
    * a = c.a;
    return tile->visible != 0 || debug;
}
static void fill_chunk(Landscape * self, Chunk * chunk) {
    if (DebugLogTerrain >= VERBOSE) {
        printf("fill chunk (%d/%d) for vbo %d.\n", chunk->x, chunk->y, chunk->vbo);
    }
    if (! chunk->v) {
        chunk->v = land_malloc(self->chunk_s * self->chunk_s * self->vertices_per_tile * self->stride * sizeof (* chunk->v));
        if (DebugLogTerrain >= VERBOSE) {
            printf("    new chunk %p: %lu bytes\n", chunk->v, self->chunk_s * self->chunk_s * 12 * 3 * 2 * sizeof (* chunk->v));
        }
    }
    GLfloat * v = chunk->v;
    int i = 0;
    //int wm = self.w - 1
#define C(X, Y) \
        get_color(self, x + (X), y + (Y), v + i + 0, v + i + 1, v + i + 2, v + i + 3); \
        i += 4;
#define WIREFRAME(vertex_number) \
        v [i++] = vertex_number;
#define T(AX, AY, BX, BY) \
        v [i++] = vx + AX; \
        v [i++] = vy + AY; \
        v [i++] = Z(AX, AY); \
        C(AX, AY); \
        WIREFRAME(0); \
        v [i++] = vx + BX; \
        v [i++] = vy + BY; \
        v [i++] = Z(BX, BY); \
        C(BX, BY); \
        WIREFRAME(1); \
        v [i++] = xm; \
        v [i++] = ym; \
        v [i++] = zm; \
        v [i++] = rm; \
        v [i++] = gm; \
        v [i++] = bm; \
        v [i++] = am; \
        WIREFRAME(2);
    bool empty = 1;
    for (int yi = 0; yi < self->chunk_s; yi += 1) {
        for (int xi = 0; xi < self->chunk_s; xi += 1) {
            int x = chunk->x + xi;
            int y = chunk->y + yi;
            float vx = x;
            float vy = y;
            float zm = (Z(0, 0) + Z(1, 0) + Z(0, 1) + Z(1, 1)) / 4;
            float xm = vx + 0.5;
            float ym = vy + 0.5;
            GLfloat r [4], g [4], b [4], a [4];
            if (get_color(self, x, y, r + 0, g + 0, b + 0, a + 0)) {
                empty = 0;
            }
            if (get_color(self, x + 1, y, r + 1, g + 1, b + 1, a + 1)) {
                empty = 0;
            }
            if (get_color(self, x + 1, y + 1, r + 2, g + 2, b + 2, a + 2)) {
                empty = 0;
            }
            if (get_color(self, x, y + 1, r + 3, g + 3, b + 3, a + 3)) {
                empty = 0;
            }
            float rm = (r [0] + r [1] + r [2] + r [3]) / 4;
            float gm = (g [0] + g [1] + g [2] + g [3]) / 4;
            float bm = (b [0] + b [1] + b [2] + b [3]) / 4;
            float am = (a [0] + a [1] + a [2] + a [3]) / 4;
            //if xi == self.chunk_s - 1: gm = 1
            //if yi == self.chunk_s - 1: gm = 1
            //if (x & wm) == 0 or (y & wm) == 0: rm = 1
            T(0, 0, 1, 0);
            T(1, 0, 1, 1);
            T(1, 1, 0, 1);
            T(0, 1, 0, 0);
        }
    }
    chunk->empty = empty;
    if (empty) {
        land_free(chunk->v);
        chunk->v = NULL;
    }
    if (DebugLogTerrain >= VERBOSE) {
        printf("    Filled %lu bytes of vertices.\n", i * sizeof (* v));
    }
}
void landscape_render(Landscape * self, Camera * cam) {
    landscape_render2(self, cam, cam);
}
void landscape_render2(Landscape * self, Camera * cam, Camera * cam2) {
    /* cam2 is only used to determine visible chunks, if it is set to a
     * fixed map position the same part of the map will always be visible
     */
    if (self->visible_chunks_w > self->max_visible_chunks_s || self->visible_chunks_h > self->max_visible_chunks_s) {
        print("ERROR: max_visible_chunks_s is too small: %d", self->max_visible_chunks_s);
        return ;
    }
    if (! self->chunks) {
        self->chunks = land_calloc(self->chunks_count * sizeof (* self->chunks));
        GLuint vbos [self->chunks_count];
        glGenBuffers(self->chunks_count, vbos);
        for (int i = 0; i < self->chunks_count; i += 1) {
            self->chunks [i].vbo = vbos [i];
            if (DebugLogTerrain >= VERBOSE) {
                printf("VBO %d: %u\n", i, vbos [i]);
            }
        }
        int s = LIGHTMAP_RESOLUTION;
        self->lightmap = land_image_new(self->w * s, self->w * s);
        land_image_set_rgba_data(self->lightmap, self->lightmap_rgba);
        land_free(self->lightmap_rgba);
        // FIXME debug only!
        //land_image_save(self.lightmap, "lightmap.png")
        glBindTexture(GL_TEXTURE_2D, land_image_opengl_texture(self->lightmap));
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    int lx = floor(cam2->p.x / self->chunk_s - self->visible_chunks_w / 2.0) * self->chunk_s;
    int ly = floor(cam2->p.y / self->chunk_s - self->visible_chunks_h / 2.0) * self->chunk_s;
    //if DebugLogTerrain >= VERBOSE:
    //    printf("%f/%f -> %d/%d (%dx%d size %d)\n", cam.p.x, cam.p.y, lx, ly,
    //        self.visible_chunks_w, self.visible_chunks_h, self.chunk_s)
    // mark position of required chunks
    bool need_any = 0;
    int got [self->max_visible_chunks_s * self->max_visible_chunks_s];
    memset(got, 0, sizeof got);
    for (int i = 0; i < self->chunks_count; i += 1) {
        Chunk * c = self->chunks + i;
        if (! c->filled) {
            continue;
        }
        if (c->x >= lx && c->x < lx + self->visible_chunks_w * self->chunk_s && c->y >= ly && c->y < ly + self->visible_chunks_h * self->chunk_s) {
            int px = (c->x - lx) / self->chunk_s;
            int py = (c->y - ly) / self->chunk_s;
            int pos = px + py * self->max_visible_chunks_s;
            got [pos] = 1;
            if (c->empty) {
                got [pos] = 2;
            }
            c->required = 1;
        }
        else {
            c->required = 0;
        }
    }
    // fill missing chunks
    for (int cy = 0; cy < self->visible_chunks_h; cy += 1) {
        for (int cx = 0; cx < self->visible_chunks_w; cx += 1) {
            if (got [cx + cy * self->max_visible_chunks_s]) {
                continue;
            }
            need_any = 1;
            int x = lx + cx * self->chunk_s;
            int y = ly + cy * self->chunk_s;
            Chunk * chunk = NULL;
            int i;
            for (i = 0; i < self->chunks_count; i += 1) {
                if (! self->chunks [i].required) {
                    chunk = self->chunks + i;
                    break;
                }
            }
            if (chunk) {
                chunk->x = x;
                chunk->y = y;
                chunk->filled = 1;
                chunk->uploaded = 0;
                chunk->required = 1;
                fill_chunk(self, chunk);
                if (DebugLogTerrain >= VERBOSE) {
                    printf("filled %d/%d (%d)\n", chunk->x, chunk->y, i);
                }
            }
            else {
                printf("no more free chunk (%dx%d needed have %d)\n", self->visible_chunks_w, self->visible_chunks_h, self->chunks_count);
            }
        }
    }
    if (need_any) {
        if (DebugLogTerrain >= INFO) {
            print("got for %d/%d:", lx, ly);
            for (int cy = 0; cy < self->visible_chunks_h; cy += 1) {
                for (int cx = 0; cx < self->visible_chunks_w; cx += 1) {
                    int g = got [cx + cy * self->visible_chunks_w];
                    printf("%c", g == 2 ? ' ' : g == 1 ? 'X' : '.');
                }
                print("");
            }
        }
    }
    if (! self->shader) {
        self->shader = land_glsl_shader_new("landscape", vertex_glsl, fragment_glsl);
    }
    GLint loc;
    GLfloat matrix [4 * 4];
    if (render_switch_shader(self->shader)) {
        //land_log_message("0 %s\n", land_opengl_error())
        loc = glGetUniformLocation(self->shader->program_object, "projection_matrix");
        land_4x4_matrix_to_gl_float(world_get_projection_matrix(), matrix);
        glUniformMatrix4fv(loc, 1, GL_FALSE, matrix);
    }
    loc = glGetUniformLocation(self->shader->program_object, "modelview_matrix");
    land_4x4_matrix_to_gl_float(camera_matrix(cam), matrix);
    glUniformMatrix4fv(loc, 1, GL_FALSE, matrix);
    loc = glGetUniformLocation(self->shader->program_object, "lightscale");
    if (loc >= 0) {
        glUniform1f(loc, 1.0 / self->w);
    }
    loc = glGetUniformLocation(self->shader->program_object, "max_height");
    if (loc >= 0) {
        glUniform1f(loc, self->max_height);
    }
    loc = glGetUniformLocation(self->shader->program_object, "use_wireframe");
    if (loc >= 0) {
        glUniform1i(loc, app_global ()->debug_wireframe);
    }
    glBindTexture(GL_TEXTURE_2D, land_image_opengl_texture(self->lightmap));
    loc = glGetUniformLocation(self->shader->program_object, "tex");
    glUniform1i(loc, 0);
    int stride = self->stride * sizeof (GLfloat);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    GLint xyz = glGetAttribLocation(self->shader->program_object, "xyz");
    glEnableVertexAttribArray(xyz);
    GLint rgba = glGetAttribLocation(self->shader->program_object, "rgba");
    glEnableVertexAttribArray(rgba);
    GLint wireframe = glGetAttribLocation(self->shader->program_object, "wireframe");
    if (wireframe >= 0) {
        glEnableVertexAttribArray(wireframe);
    }
    for (int i = 0; i < self->chunks_count; i += 1) {
        Chunk * c = self->chunks + i;
        if (! c->required) {
            continue;
        }
        if (! c->v) {
            continue;
        }
        glBindBuffer(GL_ARRAY_BUFFER, c->vbo);
        if (! c->uploaded) {
            glBufferData(GL_ARRAY_BUFFER, self->chunk_s * self->chunk_s * self->vertices_per_tile * self->stride * sizeof (* c->v), c->v, GL_STATIC_DRAW);
            c->uploaded = 1;
            if (DebugLogTerrain >= VERBOSE) {
                printf("uploaded %d/%d (%d)\n", c->x, c->y, i);
            }
        }
        glVertexAttribPointer(xyz, 3, GL_FLOAT, GL_FALSE, stride, 0);
        glVertexAttribPointer(rgba, 4, GL_FLOAT, GL_FALSE, stride, (GLfloat *) 0 + 3);
        if (wireframe >= 0) {
            glVertexAttribPointer(wireframe, 1, GL_FLOAT, GL_FALSE, stride, (GLfloat *) 0 + 7);
        }
        int n = self->chunk_s * self->chunk_s;
        glDrawArrays(GL_TRIANGLES, 0, n * self->vertices_per_tile);
    }
    glDisableVertexAttribArray(xyz);
    glDisableVertexAttribArray(rgba);
    if (wireframe >= 0) {
        glDisableVertexAttribArray(wireframe);
    }
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    render_switch_shader(NULL);
}
static LandVector _get_point(Landscape * self, Camera * cam, int x, int y) {
    float w = land_display_width();
    float h = land_display_height();
    LandFloat s = pow(2, cam->zoom);
    Tile * t = landscape_get_tile(self, x, y);
    LandVector v = land_vector(x, y, t->height);
    v = land_vector_transform(v, cam->p, cam->x, cam->y, cam->z);
    v.x *= s;
    v.y *= s;
    v.x = v.x + w / 2;
    v.y = h / 2 - v.y;
    return v;
}
void landscape_draw_grid(Landscape * self, Camera * cam) {
    float w = land_display_width();
    //float h = land_display_height()
    int ww = world_width();
    LandVector t = land_vector(0, 0, 0);
    LandFloat s = pow(2, cam->zoom);
    t = land_vector_backtransform(t, cam->p, cam->x, cam->y, cam->z);
    int mx = floor(t.x);
    int my = floor(t.y);
    int n = w / s / 2;
    int grid_spacing = n / 50;
    if (grid_spacing < 1) {
        grid_spacing = 1;
    }
    int text_spacing = n / 10;
    if (text_spacing < 1) {
        text_spacing = 1;
    }
    for (int y = - n; y < n; y += grid_spacing) {
        for (int x = - n; x < n; x += grid_spacing) {
            int lx = mx + x;
            int ly = my + y;
            LandVector v0 = _get_point(self, cam, lx, ly);
            LandVector v1 = _get_point(self, cam, lx + 1, ly);
            LandVector v2 = _get_point(self, cam, lx, ly + 1);
            LandVector v3 = _get_point(self, cam, lx + 1, ly + 1);
            land_premul(1, 0, 0, 0.1);
            land_line(v0.x, v0.y, v1.x, v1.y);
            if (x == n - 1 || grid_spacing > 1) {
                land_line(v1.x, v1.y, v3.x, v3.y);
            }
            if (y == n - 1 || grid_spacing > 1) {
                land_line(v2.x, v2.y, v3.x, v3.y);
            }
            land_line(v2.x, v2.y, v0.x, v0.y);
            if (lx % text_spacing == 0 && ly % text_spacing == 0) {
                land_premul(1, 0, 0, 1);
                land_text_pos(v0.x, v0.y);
                land_print("%d/%d", ((lx % ww) + ww) % ww, ((ly % ww) + ww) % ww);
            }
        }
    }
}
void landscape_debug_clear_chunks(Landscape * self) {
    for (int i = 0; i < self->chunks_count; i += 1) {
        Chunk * c = self->chunks + i;
        c->filled = 0;
        c->required = 0;
        c->uploaded = 0;
    }
}
void landscape_destroy(Landscape * self) {
    land_free(self->heightmap);
    if (self->chunks) {
        for (int i = 0; i < self->chunks_count; i += 1) {
            Chunk * c = self->chunks + i;
            if (c->v) {
                land_free(c->v);
            }
            glDeleteBuffers(1, & c->vbo);
        }
        land_free(self->chunks);
    }
    if (self->shader) {
        land_glsl_shader_destroy(self->shader);
    }
    if (self->lightmap) {
        land_image_destroy(self->lightmap);
    }
    if (self->adjustments) {
        land_array_destroy_with_free(self->adjustments);
    }
    land_free(self);
}
void landscape_color(Landscape * self, float x, float y, float r, float g, float b) {
    land_set_image_display(self->lightmap);
    land_render_state(LAND_WRITE_MASK, LAND_RGB_MASK);
    int s = 4;
    s *= 2;
    for (int v = - s; v < s + 1; v += 1) {
        for (int u = - s; u < s + 1; u += 1) {
            float c = 1.0 - sqrtf(u * u + v * v) / s;
            if (c < 0) {
                continue;
            }
            land_color(r * c, g * c, b * c, c);
            float lx = self->w + rwrap(x * 2 + u + 0.5, self->w * 2);
            float ly = self->w - rwrap(y * 2 + v + 0.5, self->w * 2);
            land_plot(lx, ly);
        }
    }
    land_render_state(LAND_WRITE_MASK, LAND_RGBA_MASK | LAND_DEPTH_MASK);
    land_unset_image_display();
}
void landscape_height(Landscape * self, int x, int y, float height) {
    land_set_image_display(self->lightmap);
    land_render_state(LAND_WRITE_MASK, LAND_ALPHA_MASK);
    land_blend(LAND_BLEND_SOLID);
    int s = 4;
    for (int v = - s; v < s + 1; v += 1) {
        for (int u = - s; u < s + 1; u += 1) {
            float c = 1.0 - sqrtf(u * u + v * v) / s;
            if (c < 0) {
                continue;
            }
            Tile * t = landscape_get_tile(self, x + u, y + v);
            t->height = (1 - c) * t->height + c * height;
        }
    }
    s *= 2;
    for (int v = - s; v < s + 1; v += 1) {
        for (int u = - s; u < s + 1; u += 1) {
            float h = landscape_get_height_interpolated(self, x + u / 2.0, y + v / 2.0);
            land_color(0, 0, 0, h / self->max_height);
            float lx = self->w + rwrap(x * 2 + u + 0.5, self->w * 2);
            float ly = self->w - rwrap(y * 2 + v + 0.5, self->w * 2);
            land_plot(lx, ly);
        }
    }
    land_render_state(LAND_WRITE_MASK, LAND_RGBA_MASK | LAND_DEPTH_MASK);
    land_blend(0);
    land_unset_image_display();
}
void landscape_invalidate(Landscape * self) {
    if (! self->chunks) {
        return ;
    }
    for (int i = 0; i < self->chunks_count; i += 1) {
        Chunk * c = self->chunks + i;
        c->filled = 0;
        c->uploaded = 0;
        c->required = 0;
    }
}
void landscape_make_circle_visible(Landscape * self, int cx, int cy, int rn, int rf) {
    int r = rf;
    int r2 = rn;
    bool invalidated = 0;
    for (int y = - r; y < r + 1; y += 1) {
        for (int x = - r; x < r + 1; x += 1) {
            if (x * x + y * y > r * r) {
                continue;
            }
            Tile * t = landscape_get_tile(self, cx + x, cy + y);
            int v;
            if (x * x + y * y < r2 * r2) {
                v = 255;
            }
            else {
                float d = sqrt(x * x + y * y);
                v = 255 - 255 * (d - r2) / (r - r2);
            }
            if (v > t->visible) {
                invalidated = 1;
                t->visible = v;
            }
        }
    }
    if (invalidated) {
        landscape_invalidate(self);
    }
}
void landscape_adjust(Landscape * ls, int x, int y, int s, int b, float height) {
    Adjustment * ad;
    land_alloc(ad);
    ad->x = x;
    ad->y = y;
    ad->s = s;
    ad->b = b;
    ad->height = height;
    print("adjusting landscape %d/%d %d+%d -> %.1f", x, y, s, b, height);
    land_array_add(ls->adjustments, ad);
    for (int j = - s - b; j < s + b + 1; j += 1) {
        for (int i = - s - b; i < s + b + 1; i += 1) {
            Tile * tile = landscape_get_tile(ls, x + i, y + j);
            float f = 1;
            double r = sqrt(i * i + j * j);
            if (r > s) {
                if (r > b + s) {
                    continue;
                }
                f = 1.0 - (r - s) / (float)(s + b - s);
            }
            tile->height += (height - tile->height) * f;
        }
    }
    int ur = s + b + 1;
    isles_update_normals_area(ls, x - ur, y - ur, ur + 1 + ur, ur + 1 + ur);
}
LandImage* landscape_create_visible_picture(Landscape * ls) {
    LandImage * picture = land_image_new(ls->w, ls->w);
    uint8_t * rgba = land_malloc(ls->w * ls->w * 4);
    uint8_t * p = rgba;
    for (int y = 0; y < ls->w; y += 1) {
        for (int x = 0; x < ls->w; x += 1) {
            Tile * tile = landscape_get_tile(ls, x, y);
            uint8_t c = tile->visible;
            p [0] = c;
            p [1] = c;
            p [2] = c;
            p [3] = 255;
            p += 4;
        }
    }
    land_image_set_rgba_data(picture, rgba);
    land_free(rgba);
    return picture;
}
void landscape_apply_visibility_picture(Landscape * ls, LandImage * pic) {
    uint8_t * rgba = land_malloc(ls->w * ls->w * 4);
    land_image_get_rgba_data(pic, rgba);
    uint8_t * p = rgba;
    for (int y = 0; y < ls->w; y += 1) {
        for (int x = 0; x < ls->w; x += 1) {
            Tile * tile = landscape_get_tile(ls, x, y);
            tile->visible = p [0];
            p += 4;
        }
    }
    land_free(rgba);
}
