#include "game/mesh.h"
//import landscape
    /* A mesh. It's a list of vertices.
     */
static LandVector vec(GLfloat * xyz);
static void _write_int(LandFile * f, int x);
static int _read_int(LandFile * f);
static void _write_string(LandFile * f, str s);
static char* _read_string(LandFile * f);
static void _write(Mesh * self, LandFile * f);
static void _read(Mesh * self, LandFile * f, char * section);
static void _trans(GLfloat * xyz, Land4x4Matrix * matrix);
Mesh* mesh_new(void) {
    Mesh * self;
    land_alloc(self);
    self->stride = 11;
    self->aabb = land_csg_aabb_empty();
    return self;
}
Mesh* mesh_new_textured(void) {
    Mesh * self;
    land_alloc(self);
    self->stride = 13;
    return self;
}
MeshMarker* mesh_find_marker(Mesh * self, char const * name) {
    if (! self->markers) {
        return NULL;
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(self->markers);
        for (MeshMarker * m = LandArrayIterator_item(self->markers, &__iter0__); LandArrayIterator_next(self->markers, &__iter0__); m = LandArrayIterator_item(self->markers, &__iter0__)) {
            if (land_equals(m->name, name)) {
                return m;
            }
        }
    }
    return NULL;
}
void mesh_destroy(Mesh * self) {
    if (self->vbo) {
        glDeleteBuffers(1, & self->vbo);
    }
    if (self->v) {
        land_free(self->v);
    }
    land_free(self);
}
void mesh_add_marker(Mesh * self, Object * o) {
    if (! o->name) {
        print("Error: object is no marker");
        return ;
    }
    if (! self->markers) {
        self->markers = land_array_new();
    }
    MeshMarker * marker;
    land_alloc(marker);
    marker->name = o->name;
    marker->transform = o->transform;
    land_array_add(self->markers, marker);
}
void mesh_add_object(Mesh * self, Object * o) {
    if (! o->csg) {
        return ;
    }
    // only used to set the object aabb
    land_csg_aabb_update(& o->csg->bbox, o->csg->polygons);
    int i = self->n;
    land_csg_triangles(o->csg);
    int n = self->n + land_array_count(o->csg->polygons) * 3;
    self->v = land_realloc(self->v, n * self->stride * sizeof (* self->v));
    self->n = n;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(o->csg->polygons);
        for (LandCSGPolygon * t = LandArrayIterator_item(o->csg->polygons, &__iter0__); LandArrayIterator_next(o->csg->polygons, &__iter0__); t = LandArrayIterator_item(o->csg->polygons, &__iter0__)) {
            for (int j = 0; j < 3; j += 1) {
                LandCSGVertex * vec = land_array_get_nth(t->vertices, j);
                GLfloat * vt = self->v + self->stride * i;
                vt [0] = vec->pos.x;
                vt [1] = vec->pos.y;
                vt [2] = vec->pos.z;
                vt [3] = vec->normal.x;
                vt [4] = vec->normal.y;
                vt [5] = vec->normal.z;
                Material * material = t->shared;
                if (material) {
                    // Note: values > 1 will go all the way to the shader,
                    // especially also the alpha value will be multiplied
                    // in for blending.
                    vt [6] = material->r + vec->rgba.r;
                    vt [7] = material->g + vec->rgba.g;
                    vt [8] = material->b + vec->rgba.b;
                    vt [9] = material->a + vec->rgba.a;
                }
                else {
                    vt [6] = 1;
                    vt [7] = 0;
                    vt [8] = 1;
                    vt [9] = 0.5;
                }
                vt [10] = j;
                i++;
            }
        }
    }
    self->aabb = land_csg_aabb_combine(self->aabb, o->csg->bbox);
}
void mesh_get_bounds(Mesh * self, LandFloat * x1, LandFloat * y1, LandFloat * z1, LandFloat * x2, LandFloat * y2, LandFloat * z2) {
    * x1 = * x2 = self->v [0];
    * y1 = * y2 = self->v [1];
    * z1 = * z2 = self->v [2];
    for (int i = 0; i < self->n; i += 1) {
        GLfloat * vt = self->v + self->stride * i;
        if (vt [0] < * x1) {
            * x1 = vt [0];
        }
        if (vt [1] < * y1) {
            * y1 = vt [1];
        }
        if (vt [2] < * z1) {
            * z1 = vt [2];
        }
        if (vt [0] > * x2) {
            * x2 = vt [0];
        }
        if (vt [1] > * y2) {
            * y2 = vt [1];
        }
        if (vt [2] > * z2) {
            * z2 = vt [2];
        }
    }
}
void mesh_get_bounds_if_less(Mesh * self, LandFloat * x1, LandFloat * y1, LandFloat * z1, LandFloat * x2, LandFloat * y2, LandFloat * z2, LandFloat x, LandFloat y, LandFloat z) {
    * x1 = * x2 = self->v [0];
    * y1 = * y2 = self->v [1];
    * z1 = * z2 = self->v [2];
    for (int i = 0; i < self->n; i += 1) {
        GLfloat * vt = self->v + self->stride * i;
        if (vt [0] >= x) {
            continue;
        }
        if (vt [1] >= y) {
            continue;
        }
        if (vt [2] >= z) {
            continue;
        }
        if (vt [0] < * x1) {
            * x1 = vt [0];
        }
        if (vt [1] < * y1) {
            * y1 = vt [1];
        }
        if (vt [2] < * z1) {
            * z1 = vt [2];
        }
        if (vt [0] > * x2) {
            * x2 = vt [0];
        }
        if (vt [1] > * y2) {
            * y2 = vt [1];
        }
        if (vt [2] > * z2) {
            * z2 = vt [2];
        }
    }
}
void mesh_get_bounds_if_between(Mesh * self, LandFloat * x1, LandFloat * y1, LandFloat * z1, LandFloat * x2, LandFloat * y2, LandFloat * z2, LandFloat xl, LandFloat yl, LandFloat zl, LandFloat xg, LandFloat yg, LandFloat zg) {
    * x1 = * x2 = self->v [0];
    * y1 = * y2 = self->v [1];
    * z1 = * z2 = self->v [2];
    for (int i = 0; i < self->n; i += 1) {
        GLfloat * vt = self->v + self->stride * i;
        if (vt [0] >= xg) {
            continue;
        }
        if (vt [1] >= yg) {
            continue;
        }
        if (vt [2] >= zg) {
            continue;
        }
        if (vt [0] < xl) {
            continue;
        }
        if (vt [1] < yl) {
            continue;
        }
        if (vt [2] < zl) {
            continue;
        }
        if (vt [0] < * x1) {
            * x1 = vt [0];
        }
        if (vt [1] < * y1) {
            * y1 = vt [1];
        }
        if (vt [2] < * z1) {
            * z1 = vt [2];
        }
        if (vt [0] > * x2) {
            * x2 = vt [0];
        }
        if (vt [1] > * y2) {
            * y2 = vt [1];
        }
        if (vt [2] > * z2) {
            * z2 = vt [2];
        }
    }
}
void mesh_set_origin(Mesh * self, float x, float y, float z) {
    for (int i = 0; i < self->n; i += 1) {
        GLfloat * vt = self->v + self->stride * i;
        vt [0] -= x;
        vt [1] -= y;
        vt [2] -= z;
    }
}
static LandVector vec(GLfloat * xyz) {
    return land_vector(xyz [0], xyz [1], xyz [2]);
}
void mesh_recalculate_normals(Mesh * self) {
    int s = self->stride;
    for (int i = 0; i < self->n; i += 3) {
        LandVector t0 = vec(self->v + s * (i + 0));
        LandVector t1 = vec(self->v + s * (i + 1));
        LandVector t2 = vec(self->v + s * (i + 2));
        LandVector a = land_vector_sub(t0, t1);
        LandVector b = land_vector_sub(t1, t2);
        LandVector n = land_vector_normalize(land_vector_cross(a, b));
        for (int j = 0; j < 3; j += 1) {
            GLfloat * v = self->v + s * (i + j);
            v [3] = n.x;
            v [4] = n.y;
            v [5] = n.z;
        }
    }
}
void mesh_triangle_callback(Mesh * self, void(* callback)(LandVector t0, LandVector t1, LandVector t2, void * data), void * data) {
    int s = self->stride;
    for (int i = 0; i < self->n; i += 3) {
        LandVector t0 = vec(self->v + s * (i + 0));
        LandVector t1 = vec(self->v + s * (i + 1));
        LandVector t2 = vec(self->v + s * (i + 2));
        callback(t0, t1, t2, data);
    }
}
static void _write_int(LandFile * f, int x) {
    land_file_write(f, (char *) & x, sizeof x);
}
static int _read_int(LandFile * f) {
    int x;
    land_file_read(f, (char *) & x, sizeof x);
    return x;
}
static void _write_string(LandFile * f, str s) {
    int n = strlen(s);
    _write_int(f, n);
    land_file_write(f, s, n);
}
static char* _read_string(LandFile * f) {
    int n = _read_int(f);
    char * s = land_malloc(n + 1);
    land_file_read(f, s, n);
    s [n] = 0;
    return s;
}
static void _write(Mesh * self, LandFile * f) {
    land_file_write(f, "MESH", 4);
    _write_int(f, self->n);
    _write_int(f, self->stride);
    land_file_write(f, (char *) self->v, self->n * self->stride * sizeof (* self->v));
    land_file_write(f, "MARK", 4);
    int n = self->markers ? land_array_count(self->markers) : 0;
    _write_int(f, n);
    if (n) {
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(self->markers);
            for (MeshMarker * m = LandArrayIterator_item(self->markers, &__iter0__); LandArrayIterator_next(self->markers, &__iter0__); m = LandArrayIterator_item(self->markers, &__iter0__)) {
                _write_string(f, m->name);
                land_file_write(f, (char *) & m->transform, sizeof (Land4x4Matrix));
            }
        }
    }
    land_file_write(f, "AABB", 4);
    land_file_write(f, (char *) & self->aabb.x1, sizeof (double));
    land_file_write(f, (char *) & self->aabb.y1, sizeof (double));
    land_file_write(f, (char *) & self->aabb.z1, sizeof (double));
    land_file_write(f, (char *) & self->aabb.x2, sizeof (double));
    land_file_write(f, (char *) & self->aabb.y2, sizeof (double));
    land_file_write(f, (char *) & self->aabb.z2, sizeof (double));
}
static void _read(Mesh * self, LandFile * f, char * section) {
    self->n = _read_int(f);
    self->stride = _read_int(f);
    int bytes_count = self->n * self->stride * sizeof (* self->v);
    self->v = land_malloc(bytes_count);
    land_file_read(f, (char *) self->v, bytes_count);
    land_file_read(f, section, 4);
    if (land_equals(section, "MARK")) {
        int n = _read_int(f);
        if (n) {
            self->markers = land_array_new();
            for (int i = 0; i < n; i += 1) {
                MeshMarker * m = land_calloc(sizeof (* m));
                m->name = _read_string(f);
                land_file_read(f, (char *) & m->transform, sizeof (Land4x4Matrix));
                land_array_add(self->markers, m);
            }
        }
        land_file_read(f, section, 4);
    }
    if (land_equals(section, "AABB")) {
        land_file_read(f, (char *) & self->aabb.x1, sizeof (double));
        land_file_read(f, (char *) & self->aabb.y1, sizeof (double));
        land_file_read(f, (char *) & self->aabb.z1, sizeof (double));
        land_file_read(f, (char *) & self->aabb.x2, sizeof (double));
        land_file_read(f, (char *) & self->aabb.y2, sizeof (double));
        land_file_read(f, (char *) & self->aabb.z2, sizeof (double));
        land_file_read(f, section, 4);
    }
}
int mesh_write_frames(LandFile * f, LandArray * frames) {
    int wrote = 0;
    int n = land_array_count(frames);
    land_file_write(f, "FRUT", 4);
    land_file_put32le(f, n);
    for (int i = 0; i < n; i += 1) {
        Mesh * mesh = land_array_get(frames, i);
        if (mesh) {
            _write(mesh, f);
            wrote++;
        }
    }
    return wrote;
}
LandArray* mesh_read_frames(LandFile * f) {
    LandArray * frames = NULL;
    char section [5];
    land_file_read(f, section, 4);
    if (! land_equals(section, "FRUT")) {
        return NULL;
    }
    int n = land_file_get32le(f);
    land_file_read(f, section, 4);
    for (int i = 0; i < n; i += 1) {
        if (! land_equals(section, "MESH")) {
            return NULL;
        }
        Mesh * mesh = mesh_new();
        _read(mesh, f, section);
        if (! frames) {
            frames = land_array_new();
        }
        land_array_add(frames, mesh);
    }
    return frames;
}
static void _trans(GLfloat * xyz, Land4x4Matrix * matrix) {
    LandVector v = land_vector(xyz [0], xyz [1], xyz [2]);
    LandVector v_ = land_vector_matmul(v, matrix);
    xyz [0] = v_.x;
    xyz [1] = v_.y;
    xyz [2] = v_.z;
}
void mesh_transform_positions(Mesh * self, Land4x4Matrix * matrix) {
    /* """Transform positions (not normals)."""
     */
    int s = self->stride;
    for (int i = 0; i < self->n; i += 3) {
        _trans(self->v + s * (i + 0), matrix);
    }
}
