#include "game/model.h"
static ModelRender * global_model_render;
static void cleanup(void);
static void render_mesh(Mesh * mesh, Land4x4Matrix parent_matrix);
static void cleanup(void) {
    if (global_model_render) {
        if (global_model_render->shader) {
            land_glsl_shader_destroy(global_model_render->shader);
        }
        land_free(global_model_render);
    }
}
ModelRender* model_render_singleton(void) {
    if (! global_model_render) {
        common_add_cleanup(cleanup);
        land_alloc(global_model_render);
    }
    return global_model_render;
}
#if defined(ANDROID) || defined(MACOS)
static char const * vertex_glsl = "\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "attribute vec4 xyz;\n"
    "attribute vec3 normal;\n"
    "attribute vec4 rgba;\n"
    "uniform mat4 modelview_matrix, projection_matrix;\n"
    "uniform vec3 light_direction;\n"
    "varying vec4 color;\n"
    "varying float darkness;\n"
    "void main(void) {\n"
    "    gl_Position = projection_matrix * modelview_matrix * xyz;\n"
    "    float d = dot(normal, light_direction);\n"
    "    float d1 = (1.0 + d) / 2.0;\n"
    "    float d2 = (1.0 - d) / 4.0;\n"
    "    darkness = d1 + d2;\n"
    "    color = rgba;\n"
    "}\n"
    "";
static char const * fragment_glsl = "\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "varying vec4 color;\n"
    "varying float darkness;\n"
    "\n"
    "void main(void) {\n"
    "    gl_FragColor = vec4(color.xyz * darkness * color.w, color.w);\n"
    "}\n"
    "";
#else
static char const * vertex_glsl = "\n"
    "#version 300 es // for fragment shader\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "in vec4 xyz;\n"
    "in vec3 normal;\n"
    "in vec4 rgba;\n"
    "in float wireframe;\n"
    "uniform mat4 modelview_matrix, projection_matrix;\n"
    "uniform vec3 light_direction;\n"
    "out vec4 color;\n"
    "out float darkness;\n"
    "out vec3 barycentric;\n"
    "void main(void) {\n"
    "    gl_Position = projection_matrix * modelview_matrix * xyz;\n"
    "    float d = dot(normal, light_direction);\n"
    "    float d1 = (1.0 + d) / 2.0;\n"
    "    float d2 = (1.0 - d) / 4.0;\n"
    "    darkness = d1 + d2;\n"
    "    color = rgba;\n"
    "    vec3 a[3] = vec3[](vec3(1, 0, 0), vec3(0, 1, 0), vec3(0, 0, 1));\n"
    "    barycentric = a[int(wireframe)];\n"
    "}\n"
    "";
static char const * fragment_glsl = "\n"
    "#version 300 es // for fwidth\n"
    "#ifdef GL_ES\n"
    "precision lowp float;\n"
    "#endif\n"
    "in vec4 color;\n"
    "in float darkness;\n"
    "in vec3 barycentric;\n"
    "uniform bool use_wireframe;\n"
    "uniform bool use_texture;\n"
    "uniform float invisibility;\n"
    "out vec4 out_color;\n"
    "float edgeFactor() {\n"
    "    vec3 d = fwidth(barycentric);\n"
    "    vec3 a3 = smoothstep(vec3(0.0), d*1.5, barycentric);\n"
    "    return min(min(a3.x, a3.y), a3.z);\n"
    "}\n"
    "\n"
    "void main(void) {\n"
    "    if (use_wireframe) {\n"
    "        float f = 1.0 - edgeFactor();\n"
    "        out_color = vec4(color.xyz * darkness * color.w * f,\n"
    "            color.w * f);\n"
    "    }\n"
    "    else {\n"
    "        out_color = vec4(color.xyz * darkness * invisibility * color.w, color.w);\n"
    "    }\n"
    "}\n"
    "";
#endif
void model_set_glsl(str vertex, str fragment, str name) {
    land_log_message("setting custom model shaders for %s\n", name);
    vertex_glsl = vertex;
    fragment_glsl = fragment;
}
static void render_mesh(Mesh * mesh, Land4x4Matrix parent_matrix) {
    ModelRender * g = model_render_singleton();
    Mesh * p = mesh;
    land_bind_vbo(& p->vbo, p->v, p->n * mesh->stride * sizeof (* p->v), GL_STATIC_DRAW);
    Land4x4Matrix matrix = parent_matrix;
    GLfloat gl_matrix [4 * 4];
    GLint loc = glGetUniformLocation(g->shader->program_object, "modelview_matrix");
    land_4x4_matrix_to_gl_float(matrix, gl_matrix);
    glUniformMatrix4fv(loc, 1, GL_FALSE, gl_matrix);
    GLint stride = mesh->stride * sizeof (GLfloat);
    glVertexAttribPointer(g->xyz, 3, GL_FLOAT, GL_FALSE, stride, 0);
    glVertexAttribPointer(g->normal, 3, GL_FLOAT, GL_FALSE, stride, (GLfloat *) 0 + 3);
    glVertexAttribPointer(g->rgba, 4, GL_FLOAT, GL_FALSE, stride, (GLfloat *) 0 + 6);
    if (g->wireframe >= 0) {
        glVertexAttribPointer(g->wireframe, 1, GL_FLOAT, GL_FALSE, stride, (GLfloat *) 0 + 10);
    }
    if (mesh->stride > 11) {
        if (g->uv >= 0) {
            glVertexAttribPointer(g->uv, 2, GL_FLOAT, GL_FALSE, stride, (GLfloat *) 0 + 11);
        }
    }
    glDrawArrays(GL_TRIANGLES, 0, p->n);
    land_unbind_vbo(p->vbo);
}
void model_render_object(Mesh * mesh, Land4x4Matrix matrix, Camera * cam) {
    model_render_object_light(mesh, matrix, cam, land_vector(0.75, - 1, 2));
}
void model_render_object_light(Mesh * mesh, Land4x4Matrix matrix, Camera * cam, LandVector light) {
    model_render_object_light2(mesh, matrix, world_get_projection_matrix(), cam, light);
}
void model_render_object_light2(Mesh * mesh, Land4x4Matrix matrix, Land4x4Matrix projection, Camera * cam, LandVector light) {
    model_render_object_light3(mesh, matrix, world_get_projection_matrix(), cam, light, 1.0f);
}
void model_render_object_light3(Mesh * mesh, Land4x4Matrix matrix, Land4x4Matrix projection, Camera * cam, LandVector light, float alpha) {
    ModelRender * g = model_render_singleton();
    if (! g->shader) {
        g->shader = land_glsl_shader_new("model", vertex_glsl, fragment_glsl);
    }
    //g.shader = land_glsl_shader_new_empty("model")
    //glBindAttribLocation(g.shader.program_object, 1, "wireframe") # TODO: why?
    //land_glsl_shader_set_shaders(g.shader, vertex_glsl, fragment_glsl)
    GLint loc;
    if (render_switch_shader(g->shader)) {
        GLfloat gl_matrix [4 * 4];
        glEnable(GL_CULL_FACE);
        glEnable(GL_DEPTH_TEST);
        glEnable(GL_TEXTURE_2D);
        glDepthMask(GL_TRUE);
        loc = glGetUniformLocation(g->shader->program_object, "projection_matrix");
        land_4x4_matrix_to_gl_float(projection, gl_matrix);
        glUniformMatrix4fv(loc, 1, GL_FALSE, gl_matrix);
        g->xyz = glGetAttribLocation(g->shader->program_object, "xyz");
        glEnableVertexAttribArray(g->xyz);
        g->normal = glGetAttribLocation(g->shader->program_object, "normal");
        glEnableVertexAttribArray(g->normal);
        g->rgba = glGetAttribLocation(g->shader->program_object, "rgba");
        glEnableVertexAttribArray(g->rgba);
        g->wireframe = glGetAttribLocation(g->shader->program_object, "wireframe");
        if (g->wireframe >= 0) {
            glEnableVertexAttribArray(g->wireframe);
        }
        g->uv = glGetAttribLocation(g->shader->program_object, "uv");
        if (g->uv >= 0) {
            glEnableVertexAttribArray(g->uv);
        }
        g->invisibility = glGetUniformLocation(g->shader->program_object, "invisibility");
    }
    loc = glGetUniformLocation(g->shader->program_object, "light_direction");
    light = land_vector_normalize(light);
    light = land_vector_transform(light, land_vector(0, 0, 0), cam->x, cam->y, cam->z);
    if (loc >= 0) {
        glUniform3f(loc, light.x, light.y, light.z);
    }
    loc = glGetUniformLocation(g->shader->program_object, "use_wireframe");
    if (loc >= 0) {
        glUniform1i(loc, app_global ()->debug_wireframe);
    }
    loc = glGetUniformLocation(g->shader->program_object, "use_texture");
    if (loc >= 0) {
        glUniform1i(loc, mesh->texture ? true : false);
    }
    loc = glGetUniformLocation(g->shader->program_object, "invisibility");
    if (loc >= 0) {
        glUniform1f(loc, alpha);
    }
    if (mesh->texture) {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, land_image_opengl_texture(mesh->texture));
        loc = glGetUniformLocation(g->shader->program_object, "tex");
        if (loc >= 0) {
            glUniform1i(loc, 0);
        }
    }
    render_mesh(mesh, matrix);
}
void model_render_done(void) {
    ModelRender * g = model_render_singleton();
    if (render_shader_active(g->shader)) {
        glDisableVertexAttribArray(g->xyz);
        if (g->uv >= 0) {
            glDisableVertexAttribArray(g->uv);
        }
        if (g->wireframe >= 0) {
            glDisableVertexAttribArray(g->wireframe);
        }
        glDisableVertexAttribArray(g->normal);
        glDisableVertexAttribArray(g->rgba);
        render_switch_shader(NULL);
    }
}
