#include "game/object.h"
// The material is only referenced and must exist as long as the object does.
// Ownership remains at the caller however.
Object* object_new(void) {
    Object * o;
    land_alloc(o);
    return o;
}
Object* object_clone(Object * original) {
    Object * o;
    land_alloc(o);
    o->csg = land_csg_clone(original->csg);
    return o;
}
void object_destroy(Object * self) {
    if (self->csg) {
        land_csg_destroy(self->csg);
    }
    land_free(self);
}
Object* object_sphere(int slices, int rings, Material * m) {
    Object * o = object_new();
    o->csg = csg_sphere(slices, rings, m);
    return o;
}
Object* object_spherez(int slices, int rings, Material * m) {
    Object * o = object_new();
    o->csg = csg_sphere(slices, rings, m);
    Land4x4Matrix rot90 = land_4x4_matrix_rotate(1, 0, 0, pi / 2);
    if (o->csg) {
        land_csg_transform(o->csg, rot90);
    }
    return o;
}
Object* object_dome(bool open, bool flipped, int slices, int discs, Material * m) {
    Object * o = object_new();
    o->csg = csg_dome(open, slices, discs, m);
    if (flipped) {
        land_csg_transform(o->csg, land_4x4_matrix_rotate(1, 0, 0, pi));
    }
    return o;
}
Object* object_hemi(bool open, int slices, int discs, Material * m) {
    Object * o = object_new();
    o->csg = land_csg_hemi(open, slices, discs, m);
    return o;
}
Object* object_tetrasphere(int divisions, Material * m) {
    Object * o = object_new();
    o->csg = csg_tetrasphere(divisions, m);
    return o;
}
Object* object_icosphere(int divisions, Material * m) {
    Object * o = object_new();
    o->csg = land_csg_icosphere(divisions, m);
    return o;
}
Object* object_cylinder(int segments, int discs, Material * m) {
    Object * o = object_new();
    o->csg = csg_cylinder_open(segments, discs, 0, m);
    return o;
}
Object* object_cylinder_open(int segments, int discs, Material * m) {
    Object * o = object_new();
    o->csg = csg_cylinder_open(segments, discs, 1, m);
    return o;
}
Object* object_cut_cone(int segments, float radius, Material * m) {
    Object * o = object_new();
    o->csg = csg_cut_cone_open(segments, 0, radius, m);
    return o;
}
Object* object_cone(int segments, Material * m) {
    Object * o = object_new();
    o->csg = csg_cone(segments, m);
    return o;
}
Object* object_n_pyramid(int n, Material * m) {
    Object * o = object_new();
    o->csg = csg_cut_cone_open_disced(n, 2, 0, 0, 0, m);
    return o;
}
Object* object_cut_pyramid(float x, float y, Material * m) {
    Object * o = object_new();
    o->csg = csg_cut_pyramid_open(0, x, y, m);
    return o;
}
Object* object_box(Material * m) {
    Object * o = object_new();
    o->csg = csg_cube(m);
    return o;
}
Object* object_block(int x, int y, int z, Material * m) {
    Object * o = object_new();
    o->csg = csg_block(x, y, z, 1, m);
    return o;
}
Object* object_grid(int x, int y, Material * m) {
    Object * o = object_new();
    o->csg = csg_grid(x, y, m);
    return o;
}
Object* object_regular(int n, Material * m) {
    Object * o = object_new();
    o->csg = csg_regular_polygon(n, m);
    return o;
}
Object* object_pyramid(Material * m) {
    Object * o = object_new();
    o->csg = csg_pyramid(m);
    return o;
}
Object* object_pyramid2(float x, float y, float z, Material * m) {
    Object * o = object_new();
    o->csg = csg_irregular_pyramid(x, y, z, m);
    return o;
}
Object* object_irregular_pyramid(float x, float y, float z, Material * m) {
    Object * o = object_new();
    o->csg = csg_irregular_pyramid(x, y, z, m);
    return o;
}
Object* object_irregular_cut_pyramid(float x, float y, float z, float w, float h, Material * m) {
    Object * o = object_new();
    o->csg = csg_cut_irregular_pyramid_open(0, x, y, z, w, h, m);
    return o;
}
Object* object_torus(int slices, int rings, LandFloat diameter, Material * m) {
    Object * o = object_new();
    o->csg = csg_torus(slices, rings, diameter, m);
    return o;
}
Object* object_prism(int n, Material * m) {
    /* n vertices, they all lie on a circle in the x/y plane with radius 1,
     * the first vertex is at x=0/y=1. z extends from from -1 to 1.
     */
    Object * o = object_new();
    o->csg = csg_prism(n, m);
    return o;
}
Object* object_prism2(LandFloat ax, LandFloat ay, LandFloat az, LandFloat bx, LandFloat by, LandFloat bz, LandFloat d, Material * m) {
    Object * o = object_new();
    o->csg = csg_extrude_triangle(land_vector(ax, ay, az), land_vector(bx, by, bz), d, m);
    return o;
}
Object* object_trapezoid(LandFloat x1, LandFloat x2, Material * m) {
    Object * o = object_new();
    o->csg = csg_trapezoid(x1, x2, m);
    return o;
}
Object* object_tetra(LandFloat ax, LandFloat ay, LandFloat az, LandFloat bx, LandFloat by, LandFloat bz, LandFloat cx, LandFloat cy, LandFloat cz, LandFloat dx, LandFloat dy, LandFloat dz, Material * m) {
    Object * o = object_new();
    o->csg = csg_irregular_tetrahedron(land_vector(ax, ay, az), land_vector(bx, by, bz), land_vector(cx, cy, cz), land_vector(dx, dy, dz), m);
    return o;
}
void object_subdivide(Object * o, Material * m) {
    LandCSG * newcsg = land_csg_subdivide(o->csg, m);
    land_csg_destroy(o->csg);
    o->csg = newcsg;
}
void object_shrinkwrap(Object * o, Material * m, int subdiv) {
    LandCSG * newcsg = land_csg_shrinkwrap(o->csg, m, subdiv);
    land_csg_destroy(o->csg);
    o->csg = newcsg;
}
Object* object_mark(char const * name) {
    Object * o = object_new();
    o->name = land_strdup(name);
    o->is_marker = 1;
    return o;
}
void object_transform(Object * o, Land4x4Matrix * matrix) {
    if (! o->csg) {
        o->transform = * matrix;
        return ;
    }
    land_csg_transform(o->csg, * matrix);
}
void object_union(Object * self, Object * other) {
    if (! self->csg) {
        return ;
    }
    if (! other->csg) {
        return ;
    }
    LandCSG * csg;
    if (other->inverse) {
        csg = land_csg_subtract(self->csg, other->csg);
    }
    else {
        csg = land_csg_union(self->csg, other->csg);
    }
    //print("union of %d polys with %d polys: %d polys", self.csg.polygons.count,
    //    other.csg.polygons.count, csg.polygons.count)
    land_csg_destroy(self->csg);
    self->csg = csg;
}
char* object_desc(Object * o) {
    char * r = land_strdup("[");
    if (o->csg) {
        land_append(& r, " csg %d", land_array_count(o->csg->polygons));
    }
    if (o->name) {
        land_append(& r, " marker %s", o->name);
    }
    if (o->dummy) {
        land_append(& r, " dummy");
        if (o->dummy & DummyBegin) {
            land_append(& r, " begin");
        }
        if (o->dummy & DummyUnion) {
            land_append(& r, " union");
        }
        if (o->dummy & DummyTransparent) {
            land_append(& r, " trans");
        }
        if (o->dummy & DummySeparate) {
            land_append(& r, " sep");
        }
        if (o->dummy & DummyMerge) {
            land_append(& r, " merge");
        }
        if (o->dummy & DummyRecalculateSmooth) {
            land_append(& r, " smooth");
        }
    }
    land_append(& r, " ]");
    return r;
}
