#include "script/graphics.h"
typedef struct Baking Baking;
typedef struct GraphicsState GraphicsState;
#include <land/pool.h>
#include "helper.h"
#include "application.h"
struct Baking {
    LandArray * ingredients /* list of objects to bake */;
    char * name;
};
struct GraphicsState {
    lua_State * lua;
    char * name;
    double t /* this is the "t" passed to lua */;
    double t0 /* timestamp when it was created */;
    LandArray * objects /* [Object] */;
    LandArray * materials;
    Object * clip_object;
    LandColor color;
    void * material_cb;
    Land4x4Matrix m;
    int st;
    Land4x4Matrix stack [256];
    LandArray * mesh_vertices;
    LandArray * mesh_faces;
    int mesh_recent;
    LandHash * baked /* hash of baked objects by name */;
    LandArray * bakers /* [Baking] */;
    Mesh * frames;
    int subdivisions;
};
static GraphicsState* _find_gs(lua_State * l);
static LandColor* _get_color(lua_State * l);
static Object* _dummy_new(int dummy);
static GraphicsState* _gs_new(void);
static LandArray* _current_array(GraphicsState * gs);
static Object* _merge_all_objects(LandArray * array, LandArray * trans, GraphicsState * gs);
static void _gs_frames(GraphicsState * gs);
static void gs_destroy(GraphicsState * gs);
static int _fun_color(lua_State * l);
static double _fade(double a, double b, double p);
static int _fun_fade(lua_State * l);
static int _fun_mix(lua_State * l);
static int _fun_alpha(lua_State * l);
static int _fun_rgba(lua_State * l);
static int _fun_vertex_color(lua_State * l);
static int _fun_material(lua_State * l);
static int _fun_deform(lua_State * l);
static void _recalc_normals(Object * o, int normals);
static int _fun_subdivide(lua_State * l);
static int _fun_shrinkwrap(lua_State * l);
static int _fun_push(lua_State * l);
static int _fun_pop(lua_State * l);
static int _fun_matrix_right(lua_State * l);
static int _fun_matrix_up(lua_State * l);
static int _fun_matrix_back(lua_State * l);
static int _fun_matrix_position(lua_State * l);
static int _fun_matrix_get(lua_State * l);
static int _fun_matrix_set(lua_State * l);
static void _defer(lua_State * l);
static void _create(lua_State * l, bool inverse);
static void _fun_vertex(lua_State * l);
static void _fun_face(lua_State * l);
static int _fun_mark(lua_State * l);
static int _fun_marked(lua_State * l);
static int _fun_begin_baking(lua_State * l);
static int _fun_end_baking(lua_State * l);
static int _fun_baked(lua_State * l);
static int _fun_is_baked(lua_State * l);
static int _fun_defer(lua_State * l);
static int _fun_make(lua_State * l);
static int _fun_subtract(lua_State * l);
static int _fun_begin_union(lua_State * l);
static int _fun_end_union(lua_State * l);
static int _fun_debug_gs(lua_State * l);
static int _fun_begin_merge(lua_State * l);
static int _fun_end_merge(lua_State * l);
static int _fun_begin_separate(lua_State * l);
static int _fun_end_separate(lua_State * l);
static int _fun_begin_trans(lua_State * l);
static int _fun_end_trans(lua_State * l);
static int _fun_size(lua_State * l);
static int _fun_scale(lua_State * l);
static int _fun_rotate(lua_State * l);
static int _fun_translate(lua_State * l);
static int _fun_turn(lua_State * l);
static int _fun_orient(lua_State * l);
static int _fun_lookto(lua_State * l);
static int _fun_reset(lua_State * l);
static int _fun_move(lua_State * l);
static int _fun_skew(lua_State * l);
static int _fun_cross(lua_State * l);
static int _fun_reload(lua_State * l);
static int _fun_get_subdivisions(lua_State * l);
static void thread_func(void * data);
static Model3d* _look_for_cache(str name);
static LandArray* _create_graphics_states(lua_State * l, str name, Object * clip_object);
static int lua_filter(char const * name, bool is_dir, void * data);
static GraphicsState* _find_gs(lua_State * l) {
    lua_getglobal(l, "graphics") /* [ graphics ] */;
    lua_pushstring(l, "__GraphicsState") /* [ graphics GS ] */;
    lua_rawget(l, - 2) /* [ graphics gs ] */;
    void * gs = lua_touserdata(l, - 1);
    lua_pop(l, 2) /* [] */;
    return gs;
}
static LandColor* _get_color(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    return & gs->color;
}
static Object* _dummy_new(int dummy) {
    Object * o = object_new();
    o->dummy = dummy;
    return o;
}
static GraphicsState* _gs_new(void) {
    GraphicsState * gs;
    land_alloc(gs);
    gs->materials = land_array_new();
    gs->m = land_4x4_matrix_identity();
    gs->objects = land_array_new();
    gs->bakers = land_array_new();
    gs->baked = land_hash_new();
    Application * app = app_global();
    gs->subdivisions = app->subdivisions;
    return gs;
}
static LandArray* _current_array(GraphicsState * gs) {
    LandArray * array = gs->objects;
    if (! land_array_is_empty(gs->bakers)) {
        Baking * b = land_array_get(gs->bakers, - 1);
        array = b->ingredients;
    }
    return array;
}
static Object* _merge_all_objects(LandArray * array, LandArray * trans, GraphicsState * gs) {
    Object * result = NULL;
    LandArray * stack = land_array_new();
    int separate_bits = 0;
    bool merging = 0;
    bool smooth_merged = 0;
    int inside = 0;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(array);
        for (Object * o = LandArrayIterator_item(array, &__iter0__); LandArrayIterator_next(array, &__iter0__); o = LandArrayIterator_item(array, &__iter0__)) {
            char * s = object_desc(o);
            //print("object %s (sep %x[%x])", s, separate_bits, inside)
            land_free(s);
            Object * newly_added = NULL;
            if (o->dummy & DummyBegin) {
                land_array_add(stack, result);
                result = NULL;
                inside++;
                //print("level > %d", inside)
                if (o->dummy & DummySeparate) {
                    separate_bits |= 1 << inside;
                }
                if (o->dummy & DummyMerge) {
                    merging = 1;
                    if (o->dummy & DummyRecalculateSmooth) {
                        smooth_merged = 1;
                    }
                    else {
                        smooth_merged = 0;
                    }
                }
            }
            else if (o->dummy) {
                // not begin
                inside--;
                //print("level < %d", inside)
                newly_added = result;
                result = land_array_pop(stack);
                if (o->dummy & DummyTransparent) {
                    // at end_trans there should be a single merged object which we add to trans
                    if (newly_added && trans) {
                        land_array_add(trans, newly_added);
                    }
                    newly_added = NULL;
                }
                else if (o->dummy & DummySeparate) {
                    separate_bits &= ~ (1 << (inside + 1));
                    // nothing should have been merged at the separate level
                    newly_added = NULL;
                }
                else if (o->dummy & DummyMerge) {
                    merging = 0;
                    if (newly_added) {
                        if (smooth_merged) {
                            land_csg_recalculate_smooth_normals(newly_added->csg);
                        }
                    }
                }
                else {
                    if (o->dummy & DummyRecalculateFace) {
                        land_csg_recalculate_face_normals(newly_added->csg);
                    }
                }
            }
            else if (o->csg) {
                newly_added = o;
            }
            if (newly_added) {
                //print("separate: %x %x", separate_bits, 1 << inside)
                if ((separate_bits & (1 << inside)) && trans) {
                    land_array_add(trans, newly_added);
                }
                //print("separate %d", land_array_count(newly_added.csg.polygons))
                else if (result) {
                    if (merging) {
                        land_csg_merge(result->csg, newly_added->csg);
                        newly_added->csg = NULL;
                    }
                    //print("merge %d", land_array_count(result.csg.polygons))
                    else {
                        object_union(result, newly_added);
                    }
                }
                //print("union %d", land_array_count(result.csg.polygons))
                else {
                    result = newly_added;
                }
            }
        }
    }
    //print("new %d", land_array_count(result.csg.polygons))
    land_array_destroy(stack);
    return result;
}
static void _gs_frames(GraphicsState * gs) {
    lua_getglobal(gs->lua, "graphics") /* [ graphics ] */;
    if (lua_isnil(gs->lua, - 1)) {
        print("ERROR: no graphics");
        lua_pop(gs->lua, 1) /* [] */;
        return ;
    }
    lua_pushstring(gs->lua, gs->name) /* [ graphics name ] */;
    lua_rawget(gs->lua, - 2) /* [ graphics function ] */;
    lua_remove(gs->lua, - 2) /* [ function ] */;
    if (lua_isnil(gs->lua, - 1)) {
        print("ERROR: no graphics.%s", gs->name);
        lua_pop(gs->lua, 1) /* [] */;
        return ;
    }
    lua_pushnumber(gs->lua, gs->t) /* [f, t] */;
    script_execute_params(gs->lua, gs->name, 1);
    LandArray * trans = land_array_new();
    if (DebugLuaScript) {
        print("after %.3f seconds: %d objects were created and need to be merged", land_get_time() - gs->t0, land_array_count(gs->objects));
    }
    Object * result = _merge_all_objects(gs->objects, trans, gs);
    if (gs->clip_object) {
        gs->clip_object->inverse = 1;
        if (result) {
            object_union(result, gs->clip_object);
        }
    }
    Mesh * mesh = mesh_new();
    if (result) {
        mesh_add_object(mesh, result);
        result = NULL;
    }
    if (DebugLuaScript) {
        print("after %.3f seconds: %d objects remain separated", land_get_time() - gs->t0, land_array_count(trans));
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(trans);
        for (Object * t = LandArrayIterator_item(trans, &__iter0__); LandArrayIterator_next(trans, &__iter0__); t = LandArrayIterator_item(trans, &__iter0__)) {
            if (t) {
                mesh_add_object(mesh, t);
            }
        }
    }
    land_array_destroy(trans);
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(gs->materials);
        for (Material * m = LandArrayIterator_item(gs->materials, &__iter0__); LandArrayIterator_next(gs->materials, &__iter0__); m = LandArrayIterator_item(gs->materials, &__iter0__)) {
            material_destroy(m);
        }
    }
    land_array_destroy(gs->materials);
    gs->materials = NULL;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(gs->objects);
        for (Object * o = LandArrayIterator_item(gs->objects, &__iter0__); LandArrayIterator_next(gs->objects, &__iter0__); o = LandArrayIterator_item(gs->objects, &__iter0__)) {
            if (o->is_marker) {
                mesh_add_marker(mesh, o);
            }
            object_destroy(o);
        }
    }
    land_array_destroy(gs->objects);
    //mesh_recalculate_normals(mesh)
    if (DebugLuaScript) {
        print("vertices: %d", mesh->n);
    }
    gs->frames = mesh;
}
static void gs_destroy(GraphicsState * gs) {
    land_free(gs->name);
    land_free(gs);
}
static int _fun_color(lua_State * l) {
    int n = lua_gettop(l);
    LandColor * color = _get_color(l);
    // Return previous color.
    if (1) {
        char s [1024];
        int r = round(color->r * 255);
        int g = round(color->g * 255);
        int b = round(color->b * 255);
        sprintf(s, "#%02x%02x%02x", r, g, b);
        lua_pushstring(l, s);
    }
    if (n == 1) {
        if (lua_istable(l, 1)) {
            lua_pushinteger(l, 1);
            lua_gettable(l, 1);
            lua_pushinteger(l, 2);
            lua_gettable(l, 1);
            lua_pushinteger(l, 3);
            lua_gettable(l, 1);
            lua_pushinteger(l, 4);
            lua_gettable(l, 1);
            color->r = lua_tonumber(l, - 4);
            color->g = lua_tonumber(l, - 3);
            color->b = lua_tonumber(l, - 2);
            color->a = lua_tonumber(l, - 1);
            lua_pop(l, 4);
        }
        else {
            str colors = lua_tostring(l, 1);
            if (! colors) {
                return luaL_error(l, "wrong color parameters");
            }
            if (land_equals(colors, "none")) {
                color->r = 0;
                color->g = 0;
                color->b = 0;
                color->a = 0;
            }
            else {
                * color = land_color_name(colors);
            }
        }
    }
    else if (n >= 3) {
        double r = lua_tonumber(l, 1);
        double g = lua_tonumber(l, 2);
        double b = lua_tonumber(l, 3);
        color->r = r;
        color->g = g;
        color->b = b;
        if (n == 4) {
            double a = lua_tonumber(l, 4);
            color->a = a;
        }
        else {
            color->a = 1;
        }
    }
    return 1;
}
static double _fade(double a, double b, double p) {
    return a + (b - a) * p;
}
static int _fun_fade(lua_State * l) {
    int n = lua_gettop(l);
    LandColor * color = _get_color(l);
    LandColor prev = * color;
    double p = lua_tonumber(l, n);
    lua_settop(l, n - 1);
    _fun_color(l);
    color->r = _fade(prev.r, color->r, p);
    color->g = _fade(prev.g, color->g, p);
    color->b = _fade(prev.b, color->b, p);
    color->a = _fade(prev.a, color->a, p);
    return 1;
}
static int _fun_mix(lua_State * l) {
    LandColor c1 = land_color_name(lua_tostring(l, 1));
    LandColor c2 = land_color_name(lua_tostring(l, 2));
    double p = lua_tonumber(l, 3);
    lua_pushnumber(l, _fade(c1.r, c2.r, p));
    lua_pushnumber(l, _fade(c1.g, c2.g, p));
    lua_pushnumber(l, _fade(c1.b, c2.b, p));
    lua_pushnumber(l, _fade(c1.a, c2.a, p));
    return 4;
}
static int _fun_alpha(lua_State * l) {
    int n = lua_gettop(l);
    LandColor * color = _get_color(l);
    if (n == 1) {
        double a = lua_tonumber(l, 1);
        color->r *= a;
        color->g *= a;
        color->b *= a;
        color->a = a;
    }
    return 0;
}
static int _fun_rgba(lua_State * l) {
    LandColor * color = _get_color(l);
    lua_pushnumber(l, color->r);
    lua_pushnumber(l, color->g);
    lua_pushnumber(l, color->b);
    lua_pushnumber(l, color->a);
    return 4;
}
static int _fun_vertex_color(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int pn = lua_gettop(l);
    if (pn == 1) {
        Object * o = land_array_get_nth(_current_array(gs), - 1);
        if (! o->csg) {
            return 0;
        }
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(o->csg->polygons);
            for (LandCSGPolygon * p = LandArrayIterator_item(o->csg->polygons, &__iter0__); LandArrayIterator_next(o->csg->polygons, &__iter0__); p = LandArrayIterator_item(o->csg->polygons, &__iter0__)) {
                {
                    LandArrayIterator __iter1__ = LandArrayIterator_first(p->vertices);
                    for (LandCSGVertex * v = LandArrayIterator_item(p->vertices, &__iter1__); LandArrayIterator_next(p->vertices, &__iter1__); v = LandArrayIterator_item(p->vertices, &__iter1__)) {
                        lua_pushvalue(l, 1) /* [cb] */;
                        lua_pushnumber(l, v->pos.x);
                        lua_pushnumber(l, v->pos.y);
                        lua_pushnumber(l, v->pos.z) /* [cb x y z] */;
                        int e = lua_pcall(l, 3, 4, 0);
                        if (e) {
                            printf("vertex color callback: %s\n", lua_tostring(l, - 1));
                            lua_pop(l, 1);
                            return 0;
                        }
                        LandFloat r = lua_tonumber(l, - 4);
                        LandFloat g = lua_tonumber(l, - 3);
                        LandFloat b = lua_tonumber(l, - 2);
                        LandFloat a = lua_tonumber(l, - 1);
                        lua_pop(l, 4);
                        v->rgba.r = r;
                        v->rgba.g = g;
                        v->rgba.b = b;
                        v->rgba.a = a;
                    }
                }
            }
        }
    }
    return 0;
}
static int _fun_material(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int pn = lua_gettop(l);
    if (pn == 1) {
        Object * o = land_array_get_nth(_current_array(gs), - 1);
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(o->csg->polygons);
            for (LandCSGPolygon * p = LandArrayIterator_item(o->csg->polygons, &__iter0__); LandArrayIterator_next(o->csg->polygons, &__iter0__); p = LandArrayIterator_item(o->csg->polygons, &__iter0__)) {
                LandFloat x = 0, y = 0, z = 0;
                int n = 0;
                {
                    LandArrayIterator __iter1__ = LandArrayIterator_first(p->vertices);
                    for (LandCSGVertex * v = LandArrayIterator_item(p->vertices, &__iter1__); LandArrayIterator_next(p->vertices, &__iter1__); v = LandArrayIterator_item(p->vertices, &__iter1__)) {
                        x += v->pos.x;
                        y += v->pos.y;
                        z += v->pos.z;
                        n++;
                    }
                }
                lua_pushvalue(l, 1) /* [cb] */;
                lua_pushnumber(l, x / n);
                lua_pushnumber(l, y / n);
                lua_pushnumber(l, z / n) /* [cb x y z] */;
                int e = lua_pcall(l, 3, 4, 0);
                if (e) {
                    printf("material callback: %s\n", lua_tostring(l, - 1));
                    lua_pop(l, 1);
                    return 0;
                }
                LandFloat r = lua_tonumber(l, - 4);
                LandFloat g = lua_tonumber(l, - 3);
                LandFloat b = lua_tonumber(l, - 2);
                LandFloat a = lua_tonumber(l, - 1);
                lua_pop(l, 4);
                Material * m = material_new(r, g, b, a);
                land_array_add(gs->materials, m);
                p->shared = m;
            }
        }
    }
    return 0;
}
static int _fun_deform(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int pn = lua_gettop(l);
    if (pn >= 1) {
        int normals = 0;
        if (pn >= 2) {
            normals = lua_tonumber(l, 2);
        }
        Object * o = land_array_get_nth(_current_array(gs), - 1);
        if (! o->csg) {
            return 0;
        }
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(o->csg->polygons);
            for (LandCSGPolygon * p = LandArrayIterator_item(o->csg->polygons, &__iter0__); LandArrayIterator_next(o->csg->polygons, &__iter0__); p = LandArrayIterator_item(o->csg->polygons, &__iter0__)) {
                {
                    LandArrayIterator __iter1__ = LandArrayIterator_first(p->vertices);
                    for (LandCSGVertex * v = LandArrayIterator_item(p->vertices, &__iter1__); LandArrayIterator_next(p->vertices, &__iter1__); v = LandArrayIterator_item(p->vertices, &__iter1__)) {
                        lua_pushvalue(l, 1) /* [cb] */;
                        lua_pushnumber(l, v->pos.x);
                        lua_pushnumber(l, v->pos.y);
                        lua_pushnumber(l, v->pos.z) /* [cb x y z] */;
                        int e = lua_pcall(l, 3, 3, 0);
                        if (e) {
                            printf("material callback: %s\n", lua_tostring(l, - 1));
                            lua_pop(l, 1);
                            return 0;
                        }
                        v->pos.x = lua_tonumber(l, - 3);
                        v->pos.y = lua_tonumber(l, - 2);
                        v->pos.z = lua_tonumber(l, - 1);
                        lua_pop(l, 3);
                    }
                }
            }
        }
        _recalc_normals(o, normals);
    }
    return 0;
}
static void _recalc_normals(Object * o, int normals) {
    if (normals == 1) {
        // recalculate face normals
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(o->csg->polygons);
            for (LandCSGPolygon * p = LandArrayIterator_item(o->csg->polygons, &__iter0__); LandArrayIterator_next(o->csg->polygons, &__iter0__); p = LandArrayIterator_item(o->csg->polygons, &__iter0__)) {
                land_csg_polygon_recalculate_normal(p);
            }
        }
    }
    if (normals == 2) {
        // use sphere normals
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(o->csg->polygons);
            for (LandCSGPolygon * p = LandArrayIterator_item(o->csg->polygons, &__iter0__); LandArrayIterator_next(o->csg->polygons, &__iter0__); p = LandArrayIterator_item(o->csg->polygons, &__iter0__)) {
                {
                    LandArrayIterator __iter1__ = LandArrayIterator_first(p->vertices);
                    for (LandCSGVertex * v = LandArrayIterator_item(p->vertices, &__iter1__); LandArrayIterator_next(p->vertices, &__iter1__); v = LandArrayIterator_item(p->vertices, &__iter1__)) {
                        v->normal = land_vector_normalize(v->pos);
                    }
                }
            }
        }
    }
    if (normals == 3) {
        // recalculate smooth normals
        land_csg_recalculate_smooth_normals(o->csg);
    }
}
static int _fun_subdivide(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    Material * m = material_new(gs->color.r, gs->color.g, gs->color.b, gs->color.a);
    land_array_add(gs->materials, m);
    Object * o = land_array_get_nth(_current_array(gs), - 1);
    object_subdivide(o, m);
    return 0;
}
// normals:
// 0 (default)
// 1 recalculate face normals
// 2 recalculate spherical normals
// 3 recalculate averaged vertex normals (smooth)
static int _fun_shrinkwrap(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    int subdiv = 0;
    int recalc = 0;
    if (n >= 1) {
        subdiv = lua_tointeger(l, 1);
    }
    if (n == 2) {
        recalc = lua_tointeger(l, 2);
    }
    Material * m = material_new(gs->color.r, gs->color.g, gs->color.b, gs->color.a);
    land_array_add(gs->materials, m);
    Object * o = land_array_get_nth(_current_array(gs), - 1);
    object_shrinkwrap(o, m, subdiv);
    _recalc_normals(o, recalc);
    return 0;
}
static int _fun_push(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    if (gs->st == 256) {
        print("ERROR: matrix stack overflow");
        return 0;
    }
    gs->stack [gs->st++] = gs->m;
    return 0;
}
static int _fun_pop(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    if (gs->st == 0) {
        print("ERROR: matrix stack underflow");
        return 0;
    }
    gs->m = gs->stack [--gs->st];
    return 0;
}
static int _fun_matrix_right(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    LandVector v = land_4x4_matrix_get_right(& gs->m);
    lua_pushnumber(l, v.x);
    lua_pushnumber(l, v.y);
    lua_pushnumber(l, v.z);
    return 3;
}
static int _fun_matrix_up(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    LandVector v = land_4x4_matrix_get_up(& gs->m);
    lua_pushnumber(l, v.x);
    lua_pushnumber(l, v.y);
    lua_pushnumber(l, v.z);
    return 3;
}
static int _fun_matrix_back(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    LandVector v = land_4x4_matrix_get_back(& gs->m);
    lua_pushnumber(l, v.x);
    lua_pushnumber(l, v.y);
    lua_pushnumber(l, v.z);
    return 3;
}
static int _fun_matrix_position(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    LandVector v = land_4x4_matrix_get_position(& gs->m);
    lua_pushnumber(l, v.x);
    lua_pushnumber(l, v.y);
    lua_pushnumber(l, v.z);
    LandBuffer * b = land_4x4_matrix_to_string(& gs->m);
    char * s = land_buffer_finish(b);
    //print(s)
    land_free(s);
    return 3;
}
static int _fun_matrix_get(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    for (int i = 0; i < 16; i += 1) {
        lua_pushnumber(l, gs->m.v [i]);
    }
    return 16;
}
static int _fun_matrix_set(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    for (int i = 0; i < 16; i += 1) {
        gs->m.v [i] = lua_tonumber(l, 1 + i);
    }
    return 0;
}
#define N(i) lua_tonumber(l, i)
static void _defer(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    char const * shape = lua_tostring(l, 1);
    int n = lua_gettop(l);
    Object * o = NULL;
    Material * m = material_new(gs->color.r, gs->color.g, gs->color.b, gs->color.a);
    land_array_add(gs->materials, m);
    if (DebugLuaScript) {
        print("making %s", shape);
    }
    int u = 8 * gs->subdivisions;
    int v = 8 * gs->subdivisions;
    if (land_equals(shape, "icosphere")) {
        u = gs->subdivisions;
    }
    if (land_equals(shape, "torus")) {
        LandFloat d = 0.5;
        if (n >= 2) {
            d = lua_tonumber(l, 2);
        }
        if (n >= 3) {
            u = lua_tointeger(l, 3) * gs->subdivisions;
        }
        if (n >= 4) {
            v = lua_tointeger(l, 4) * gs->subdivisions;
        }
        o = object_torus(u, v, d, m);
    }
    else {
        if (n >= 2) {
            u = lua_tointeger(l, 2) * gs->subdivisions;
        }
        if (n >= 3) {
            v = lua_tointeger(l, 3) * gs->subdivisions;
        }
    }
    if (land_equals(shape, "sphere")) {
        // note: does not work very well with deformations, our BSP
        // code is not very good with near-coplanar faces
        o = object_sphere(u, v, m);
    }
    if (land_equals(shape, "spherez")) {
        o = object_spherez(u, v, m);
    }
    if (land_equals(shape, "dome")) {
        o = object_dome(0, 0, u, v, m);
    }
    if (land_equals(shape, "dome_open")) {
        o = object_dome(1, 0, u, v, m);
    }
    if (land_equals(shape, "domeflipped")) {
        o = object_dome(0, 1, u, v, m);
    }
    if (land_equals(shape, "domeflipped_open")) {
        o = object_dome(1, 1, u, v, m);
    }
    if (land_equals(shape, "hemi")) {
        o = object_hemi(0, u, v, m);
    }
    if (land_equals(shape, "hemi_open")) {
        o = object_hemi(1, u, v, m);
    }
    if (land_equals(shape, "tetrasphere")) {
        o = object_tetrasphere(u, m);
    }
    if (land_equals(shape, "icosphere")) {
        o = object_icosphere(u, m);
    }
    if (land_equals(shape, "ball")) {
        o = object_sphere(u, v, m);
    }
    if (land_equals(shape, "tube")) {
        o = object_cylinder(u, v, m);
    }
    if (land_equals(shape, "tube_open")) {
        o = object_cylinder_open(u, v, m);
    }
    if (land_equals(shape, "rod")) {
        o = object_cylinder(u, 1, m);
    }
    if (land_equals(shape, "cylinder")) {
        o = object_cylinder(u, 1, m);
    }
    if (land_equals(shape, "cut_cone")) {
        o = object_cut_cone(u, N(3), m);
    }
    if (land_equals(shape, "cone")) {
        o = object_cone(u, m);
    }
    if (land_equals(shape, "box")) {
        if (n >= 4) {
            int x = u;
            int y = v;
            int z = lua_tointeger(l, 4) * gs->subdivisions;
            o = object_block(x, y, z, m);
        }
        else {
            o = object_box(m);
        }
    }
    if (land_equals(shape, "npyramid")) {
        o = object_n_pyramid(N(2), m);
    }
    if (land_equals(shape, "pyramid")) {
        if (n == 2) {
            o = object_cut_pyramid(N(2), N(2), m);
        }
        else if (n == 3) {
            o = object_cut_pyramid(N(2), N(3), m);
        }
        else if (n == 6) {
            if (N(5) == 0 && N(6) == 0) {
                o = object_irregular_pyramid(N(2), N(3), N(4), m);
            }
            else {
                o = object_irregular_cut_pyramid(N(2), N(3), N(4), N(5), N(6), m);
            }
        }
        else {
            o = object_pyramid(m);
        }
    }
    if (land_equals(shape, "tetra")) {
        o = object_cone(3, m);
    }
    if (land_equals(shape, "simplex")) {
        o = object_tetra(N(2), N(3), N(4), N(5), N(6), N(7), N(8), N(9), N(10), N(11), N(12), N(13), m);
    }
    // Triangular prism:
    // For all sides and corners:
    // h = 1.5
    // s = sqrt(3)
    // r = 1
    if (land_equals(shape, "prism")) {
        int sides = 3;
        if (n == 2) {
            sides = lua_tointeger(l, 2);
        }
        if (n == 1 || n == 2) {
            o = object_prism(sides, m);
        }
        else if (n == 8) {
            o = object_prism2(N(2), N(3), N(4), N(5), N(6), N(7), N(8), m);
        }
    }
    if (land_equals(shape, "trapezoid")) {
        o = object_trapezoid(N(2), N(3), m);
    }
    if (land_equals(shape, "mesh")) {
        o = object_new();
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(gs->mesh_faces);
            for (LandCSGPolygon * p = LandArrayIterator_item(gs->mesh_faces, &__iter0__); LandArrayIterator_next(gs->mesh_faces, &__iter0__); p = LandArrayIterator_item(gs->mesh_faces, &__iter0__)) {
                p->shared = m;
            }
        }
        o->csg = land_csg_new_from_polygons(gs->mesh_faces);
        land_array_destroy_with_free(gs->mesh_vertices);
        gs->mesh_vertices = NULL;
        gs->mesh_faces = NULL;
        gs->mesh_recent = 0;
    }
    if (o) {
        land_array_add(_current_array(gs), o);
    }
    else {
        luaL_error(l, "Unknown shape %s", shape);
    }
}
static void _create(lua_State * l, bool inverse) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    if (n != 0) {
        _defer(l);
    }
    Object * o = land_array_get_nth(_current_array(gs), - 1);
    if (o && o->csg) {
        object_transform(o, & gs->m);
        o->inverse = inverse;
    }
}
static void _fun_vertex(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    if (n != 6) {
        luaL_error(l, "Vertex format is x, y, z, nx, ny, nz");
        return ;
    }
    LandCSGVertex * v = land_csg_vertex_new(land_vector(N(1), N(2), N(3)), land_vector(N(4), N(5), N(6)));
    if (! gs->mesh_vertices) {
        gs->mesh_vertices = land_array_new();
    }
    land_array_add(gs->mesh_vertices, v);
    gs->mesh_recent++;
}
static void _fun_face(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    LandArray * vs = land_array_new();
    LandArray * mesh = gs->mesh_vertices;
    int count = land_array_count(mesh);
    int recent = gs->mesh_recent;
    if (n == 0) {
        for (int i = 0; i < recent; i += 1) {
            LandCSGVertex * v = land_array_get(mesh, count - recent + i);
            land_array_add(vs, land_csg_vertex_new(v->pos, v->normal));
        }
        gs->mesh_recent = 0;
    }
    else {
        for (int i = 0; i < n; i += 1) {
            LandCSGVertex * v = land_array_get(mesh, i);
            land_array_add(vs, land_csg_vertex_new(v->pos, v->normal));
        }
    }
    LandCSGPolygon * f = land_csg_polygon_new(vs, NULL);
    if (! gs->mesh_faces) {
        gs->mesh_faces = land_array_new();
    }
    land_array_add(gs->mesh_faces, f);
}
static int _fun_mark(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    char const * name = lua_tostring(l, 1);
    Object * o = object_mark(name);
    object_transform(o, & gs->m);
    land_array_add(gs->objects, o) /* TODO: is this ok when baking? */;
    return 0;
}
static int _fun_marked(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    str name = lua_tostring(l, 1);
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(gs->objects);
        for (Object * o = LandArrayIterator_item(gs->objects, &__iter0__); LandArrayIterator_next(gs->objects, &__iter0__); o = LandArrayIterator_item(gs->objects, &__iter0__)) {
            // TODO: is this ok when baking?
            if (land_equals(name, o->name)) {
                gs->m = o->transform;
                break;
            }
        }
    }
    return 0;
}
static int _fun_begin_baking(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    _fun_push(l);
    _fun_reset(l);
    Baking * b;
    land_alloc(b);
    land_array_add(gs->bakers, b) /* we support nested baking o_O */;
    if (n == 1) {
        b->name = land_strdup(lua_tostring(l, 1));
    }
    b->ingredients = land_array_new();
    return 0;
}
static int _fun_end_baking(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    Baking * b = land_array_pop(gs->bakers);
    LandArray * trans = land_array_new();
    print("baking %d ingredients for %s", land_array_count(b->ingredients), b->name);
    Object * baked = _merge_all_objects(b->ingredients, trans, gs);
    land_array_destroy(b->ingredients);
    print("    %d were separate", land_array_count(trans));
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(trans);
        for (Object * t = LandArrayIterator_item(trans, &__iter0__); LandArrayIterator_next(trans, &__iter0__); t = LandArrayIterator_item(trans, &__iter0__)) {
            if (t) {
                if (! baked) {
                    baked = t;
                }
                else {
                    land_csg_merge(baked->csg, t->csg);
                    t->csg = NULL;
                    object_destroy(t);
                }
            }
        }
    }
    land_array_destroy(trans);
    if (baked) {
        //if baked.name:
        baked->name = b->name;
        land_hash_insert(gs->baked, baked->name, baked);
    }
    //else:
    //    land_array_add(_current_array(gs), baked)
    else {
        print("nothing baked (?)");
    }
    land_free(b);
    _fun_pop(l);
    return 0;
}
static int _fun_baked(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    str name = lua_tostring(l, 1);
    Object * o = land_hash_get(gs->baked, name);
    if (o) {
        Object * baked = object_clone(o);
        object_transform(baked, & gs->m);
        land_array_add(_current_array(gs), baked);
    }
    return 0;
}
static int _fun_is_baked(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    str name = lua_tostring(l, 1);
    bool has = land_hash_has(gs->baked, name);
    print("is_baked(%s) -> %d", name, has);
    lua_pushboolean(l, has);
    return 1;
}
static int _fun_defer(lua_State * l) {
    _defer(l);
    return 0;
}
static int _fun_make(lua_State * l) {
    _create(l, 0);
    return 0;
}
static int _fun_subtract(lua_State * l) {
    _create(l, 1);
    return 0;
}
// normals:
// 0 (default)
// 1 recalculate face normals
// 2 recalculate spherical normals
// 3 recalculate averaged vertex normals (smooth)
static int _fun_begin_union(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int flags = DummyBegin | DummyUnion;
    land_array_add(_current_array(gs), _dummy_new(flags));
    return 0;
}
static int _fun_end_union(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int flags = DummyUnion;
    int n = lua_gettop(l);
    if (n == 1) {
        int kind = lua_tointeger(l, 1);
        if (kind == 1) {
            flags |= DummyRecalculateFace;
        }
    }
    land_array_add(_current_array(gs), _dummy_new(flags));
    return 0;
}
static int _fun_debug_gs(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    print("____");
    print("name: %s", gs->name);
    print("objects: %d", land_array_count(gs->objects));
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(gs->objects);
        for (Object * o = LandArrayIterator_item(gs->objects, &__iter0__); LandArrayIterator_next(gs->objects, &__iter0__); o = LandArrayIterator_item(gs->objects, &__iter0__)) {
            if (o->is_marker) {
                print("    marker: %s", o->name);
            }
            else if (o->dummy) {
                printf("    dummy:");
                if (o->dummy & DummyBegin) {
                    printf(" begin");
                }
                if (o->dummy & DummyUnion) {
                    printf(" union");
                }
                if (o->dummy & DummyTransparent) {
                    printf(" trans");
                }
                if (o->dummy & DummySeparate) {
                    printf(" sep");
                }
                if (o->dummy & DummyMerge) {
                    printf(" merge");
                }
                if (o->dummy & DummyRecalculateSmooth) {
                    printf(" smooth");
                }
                if (o->dummy & DummyRecalculateFace) {
                    printf(" norm");
                }
                print("");
            }
            else {
                print("    csg: %d polygons%s", land_array_count(o->csg->polygons), o->inverse ? " inverse" : "");
            }
        }
    }
    print("materials: %d", land_array_count(gs->materials));
    if (gs->mesh_vertices) {
        print("vertices: %d", land_array_count(gs->mesh_vertices));
    }
    if (gs->mesh_faces) {
        print("faces: %d", land_array_count(gs->mesh_faces));
    }
    print("recent: %d", gs->mesh_recent);
    print("baked: %d", land_hash_count(gs->baked));
    print("bakers: %d", land_array_count(gs->bakers));
    if (gs->frames) {
        print("frames: %d", gs->frames->n);
    }
    print("subdivisions: %d", gs->subdivisions);
    return 0;
}
static int _fun_begin_merge(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int dummy = DummyBegin | DummyMerge;
    int kind = lua_tointeger(l, 1);
    if (kind == 3) {
        dummy |= DummyRecalculateSmooth;
    }
    land_array_add(_current_array(gs), _dummy_new(dummy));
    return 0;
}
static int _fun_end_merge(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    land_array_add(_current_array(gs), _dummy_new(DummyMerge));
    return 0;
}
static int _fun_begin_separate(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    land_array_add(_current_array(gs), _dummy_new(DummyBegin | DummySeparate));
    return 0;
}
static int _fun_end_separate(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    land_array_add(_current_array(gs), _dummy_new(DummySeparate));
    return 0;
}
static int _fun_begin_trans(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    land_array_add(_current_array(gs), _dummy_new(DummyBegin | DummyTransparent));
    return 0;
}
static int _fun_end_trans(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    land_array_add(_current_array(gs), _dummy_new(DummyTransparent));
    return 0;
}
static int _fun_size(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    double x = lua_tonumber(l, 1);
    double y = n > 1 ? lua_tonumber(l, 2) : x;
    double z = n > 2 ? lua_tonumber(l, 3) : y;
    Land4x4Matrix m = land_4x4_matrix_scale(x, y, z);
    gs->m = land_4x4_matrix_mul(gs->m, m);
    return 0;
}
static int _fun_scale(lua_State * l) {
    return _fun_size(l);
}
static int _fun_rotate(lua_State * l) {
    return _fun_turn(l);
}
static int _fun_translate(lua_State * l) {
    return _fun_move(l);
}
static int _fun_turn(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    double x = lua_tonumber(l, 1);
    double y = lua_tonumber(l, 2);
    double z = lua_tonumber(l, 3);
    if (n == 4) {
        double a = lua_tonumber(l, 4);
        Land4x4Matrix r = land_4x4_matrix_rotate(x, y, z, pi * a);
        gs->m = land_4x4_matrix_mul(gs->m, r);
    }
    else {
        Land4x4Matrix rx = land_4x4_matrix_rotate(1, 0, 0, pi * x);
        Land4x4Matrix ry = land_4x4_matrix_rotate(0, 1, 0, pi * y);
        Land4x4Matrix rz = land_4x4_matrix_rotate(0, 0, 1, pi * z);
        gs->m = land_4x4_matrix_mul(gs->m, rx);
        gs->m = land_4x4_matrix_mul(gs->m, ry);
        gs->m = land_4x4_matrix_mul(gs->m, rz);
    }
    return 0;
}
static int _fun_orient(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    if (n == 0) {
        for (int j = 0; j < 12; j += 1) {
            if (j % 4 < 3) {
                gs->m.v [j] = j % 4 == (int)(j / 4) ? 1 : 0;
            }
        }
    }
    else if (n == 1) {
        int i = lua_tointeger(l, 1);
        Land4x4Matrix m = gs->stack [gs->st - i];
        for (int j = 0; j < 12; j += 1) {
            if (j % 4 < 3) {
                gs->m.v [j] = m.v [j];
            }
        }
    }
    return 0;
}
static int _fun_lookto(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    int n = lua_gettop(l);
    if (n == 3) {
        double x = lua_tonumber(l, 1);
        double y = lua_tonumber(l, 2);
        double z = lua_tonumber(l, 3);
        LandVector p = land_vector(0, 0, 0);
        LandVector b = land_vector_normalize(land_vector(- x, - y, - z));
        LandVector u = land_vector(0, 1, 0);
        LandVector r = land_vector_cross(u, b);
        // if we are looking up, orient along z
        if (land_vector_norm(r) < 0.1) {
            u = land_vector(0, 0, 1);
            r = land_vector_cross(u, b);
        }
        r = land_vector_normalize(r);
        u = land_vector_cross(b, r);
        Land4x4Matrix m = land_4x4_matrix_from_vectors(& p, & r, & u, & b);
        gs->m = land_4x4_matrix_mul(gs->m, m);
    }
    return 0;
}
static int _fun_reset(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    gs->m = land_4x4_matrix_identity();
    return 0;
}
static int _fun_move(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    double x = lua_tonumber(l, 1);
    double y = lua_tonumber(l, 2);
    double z = lua_tonumber(l, 3);
    Land4x4Matrix m = land_4x4_matrix_translate(x, y, z);
    gs->m = land_4x4_matrix_mul(gs->m, m);
    return 0;
}
static int _fun_skew(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    double xy = 0, xz = 0;
    double yx = 0, yz = 0;
    double zx = 0, zy = 0;
    int n = lua_gettop(l);
    if (n >= 1) {
        xy = lua_tonumber(l, 1);
    }
    if (n >= 2) {
        xz = lua_tonumber(l, 2);
    }
    if (n >= 3) {
        yx = lua_tonumber(l, 3);
    }
    if (n >= 4) {
        yz = lua_tonumber(l, 4);
    }
    if (n >= 5) {
        zx = lua_tonumber(l, 5);
    }
    if (n >= 6) {
        zy = lua_tonumber(l, 6);
    }
    Land4x4Matrix m = land_4x4_matrix_skew(xy, xz, yx, yz, zx, zy);
    gs->m = land_4x4_matrix_mul(gs->m, m);
    return 0;
}
static int _fun_cross(lua_State * l) {
    LandVector a = land_vector(N(1), N(2), N(3));
    LandVector b = land_vector(N(4), N(5), N(6));
    LandVector c = land_vector_cross(a, b);
    lua_pushnumber(l, c.x);
    lua_pushnumber(l, c.y);
    lua_pushnumber(l, c.z);
    return 3;
}
static int _fun_reload(lua_State * l) {
    char const * name = lua_tostring(l, 1);
    char * cwd = land_get_current_directory();
    LandBuffer * b = land_buffer_new();
    land_buffer_cat(b, cwd);
    if (! land_ends_with(cwd, "/")) {
        land_buffer_cat(b, "/");
    }
    land_free(cwd);
    land_buffer_cat(b, name);
    char * s = land_buffer_finish(b);
    script_reload(l, s);
    land_free(s);
    return 0;
}
static int _fun_get_subdivisions(lua_State * l) {
    GraphicsState * gs = _find_gs(l);
    lua_pushinteger(l, gs->subdivisions);
    return 1;
}
static void thread_func(void * data) {
    GraphicsState * gs = data;
    _gs_frames(gs);
}
// [f]
bool script_have_graphics(lua_State * l) {
    return ! lua_isnil(l, - 1);
}
void script_graphics_remove_cache(str name) {
    char * path = app_get_cache_path(name);
    land_file_remove(path);
}
static Model3d* _look_for_cache(str name) {
    Application * app = app_global();
    if (app->use_cache) {
        char * path = app_get_cache_path(name);
        // TODO: can we somehow use the filetime to guess when the cache
        // would be outdated?
        //uint64_t t = land_file_time(path)
        LandArray * frames = NULL;
        LandFile * f = land_file_new(path, "rb");
        if (f) {
            frames = mesh_read_frames(f);
            land_file_destroy(f);
        }
        if (frames) {
            int fn = land_array_count(frames);
            if (DebugLogModels) {
                printf("Retrieved %s (%d frame%s) from cache %s.\n", name, fn, fn == 1 ? "" : "s", path);
            }
            land_free(path);
            Model3d * model3d = model3d_new(name);
            model3d->frames = frames;
            return model3d;
        }
        land_free(path);
    }
    return NULL;
}
static LandArray* _create_graphics_states(lua_State * l, str name, Object * clip_object) {
    int fn = script_graphics_framecount(l, name);
    if (DebugLogModels) {
        printf("Rendering %s (%d frame%s).\n", name, fn, fn == 1 ? "" : "s");
    }
    LandArray * graphics_states = land_array_new();
    for (int i = 0; i < fn; i += 1) {
        GraphicsState * gs = _gs_new();
        land_array_add(graphics_states, gs);
        gs->clip_object = clip_object;
        gs->lua = script_init_thread(1 + i);
        gs->name = land_strdup(name);
        gs->t0 = land_get_time();
        // TODO: should just explicitly register all graphics related functions here
        script_register_same_functions_as(gs->lua, l);
        // TODO: should just rely on require() calls in lua code
        script_load_same_files_as(gs->lua, l);
        // set graphics.__GraphicsState to the per-thread gs pointer
        lua_getglobal(gs->lua, "graphics") /* [ graphics ] */;
        lua_pushstring(gs->lua, "__GraphicsState") /* [ graphics GS ] */;
        lua_pushlightuserdata(gs->lua, gs) /* [ graphics GS gs ] */;
        lua_rawset(gs->lua, - 3) /* [ graphics ] */;
        lua_pop(gs->lua, 1);
        double t;
        if (fn % 2 == 0) {
            t = - 1 + 2.0 * i / fn;
        }
        else {
            t = fn > 1 ? - 1.0 + 2.0 * i / (fn - 1.0) : 0;
        }
        gs->t = t;
    }
    return graphics_states;
}
Model3d* create_3d_model(str name, LandArray * states) {
    Model3d * model3d = model3d_new(name);
    int fn = land_array_count(states);
    LandThread * threads [fn];
    for (int i = 0; i < fn; i += 1) {
        GraphicsState * gs = land_array_get(states, i);
        if (DebugLuaScript) {
            print("starting thread %d (%d subdivisions)", i, gs->subdivisions);
        }
        threads [i] = land_thread_new(thread_func, gs);
    }
    LandArray * frames = land_array_new();
    for (int i = 0; i < fn; i += 1) {
        land_thread_destroy(threads [i]);
        GraphicsState * gs = land_array_get(states, i);
        land_array_add(frames, gs->frames);
        gs_destroy(gs);
    }
    Application * app = app_global();
    if (app->use_cache && frames) {
        char * path = app_get_cache_path(name);
        if (DebugLuaScript) {
            print("Caching %s (%d frame%s) to %s.", name, fn, fn == 1 ? "" : "s", path);
        }
        LandFile * f = land_file_new(path, "wb");
        if (f) {
            int n = mesh_write_frames(f, frames);
            land_file_destroy(f);
            if (n == 0) {
                land_file_remove(path);
                print("    failed caching any frames");
            }
        }
        land_free(path);
    }
    model3d->frames = frames;
    return model3d;
}
Model3d* script_graphics(lua_State * l, char const * name) {
    return script_graphics_clip_box(name, NULL);
}
Model3d* script_graphics_load(char const * name) {
    return script_graphics_clip_box(name, NULL);
}
Model3d* script_graphics_clip_box(char const * name, Object * clip_object) {
    double t0 = land_get_time();
    Model3d * model3d = _look_for_cache(name);
    if (model3d) {
        return model3d;
    }
    lua_State * l = script_init();
    LandArray * states = _create_graphics_states(l, name, clip_object);
    model3d = create_3d_model(name, states);
    model3d->render_time = land_get_time() - t0;
    return model3d;
}
int script_graphics_framecount(lua_State * l, char const * name) {
    char sname [1024];
    strcpy(sname, name);
    strcat(sname, ".animation");
    lua_getglobal(l, "graphics") /* [ graphics ] */;
    if (lua_isnil(l, - 1)) {
        lua_pop(l, 1) /* [] */;
        return 0;
    }
    lua_pushstring(l, sname) /* [ graphics sname ] */;
    lua_rawget(l, - 2) /* [ graphics static ] */;
    char const * animation = lua_tostring(l, - 1);
    if (DebugLogModels) {
        print("Animation %s: %s", sname, animation ? animation : "(default 5 frames)");
    }
    int c = 5;
    if (animation) {
        if (land_equals(animation, "static")) {
            c = 1;
        }
        if (land_equals(animation, "around")) {
            c = 32;
        }
    }
    lua_pop(l, 2) /* [ ] */;
    return c;
}
static int lua_filter(char const * name, bool is_dir, void * data) {
    if (is_dir) {
        return 2;
    }
    if (! land_ends_with(name, ".lua")) {
        return 0;
    }
    return 3;
}
void graphics_reload_all(void) {
    graphics_reload_all_from(NULL);
}
void graphics_reload_all_from(char const * subdir) {
    double t0 = land_get_time();
    char directory [256] = "lua/graphics";
    if (subdir) {
        strcat(directory, "/");
        strcat(directory, subdir);
    }
    lua_State * l = script_init();
    LandArray * a = land_filelist(directory, lua_filter, LAND_FULL_PATH, NULL);
    if (! a) {
        print("Error: Scripts not found in %s.", directory);
        return ;
    }
    int count = 0;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(a);
        for (char * name = LandArrayIterator_item(a, &__iter0__); LandArrayIterator_next(a, &__iter0__); name = LandArrayIterator_item(a, &__iter0__)) {
            if (script_load(l, name)) {
                script_execute(l, name);
                count++;
            }
            else {
                lua_pop(l, 1);
            }
            land_free(name);
        }
    }
    land_array_destroy(a);
    double dt = land_get_time() - t0;
    print("graphics_reload_all_from %s loaded %d files in %.3f seconds", subdir, count, dt);
}
void graphics_load_all(void) {
    graphics_load_all_from(NULL);
}
void graphics_load_all_from(char const * subdir) {
    lua_State * l = script_init();
    add(color);
    add(fade);
    add(push);
    add(pop);
    add(make);
    add(size);
    add(scale);
    add(turn);
    add(rotate);
    add(move);
    add(translate);
    add(reload);
    add(subtract);
    add(begin_union);
    add(end_union);
    add(begin_trans);
    add(end_trans);
    add(material);
    add(deform);
    add(defer);
    add(mark);
    add(vertex_color);
    add(orient);
    add(alpha);
    add(rgba);
    add(lookto);
    add(cross);
    add(marked);
    add(skew);
    add(begin_separate);
    add(end_separate);
    add(vertex);
    add(face);
    add(begin_baking);
    add(end_baking);
    add(baked);
    add(is_baked);
    add(reset);
    add(begin_merge);
    add(end_merge);
    add(matrix_right);
    add(matrix_up);
    add(matrix_back);
    add(matrix_position);
    add(matrix_get);
    add(matrix_set);
    add(get_subdivisions);
    add(mix);
    add(subdivide);
    add(debug_gs);
    add(shrinkwrap);
    graphics_reload_all_from(subdir);
}
