#include "script/script.h"
typedef struct Script Script;
#include "setjmp.h"
#include "script/helper.h"
lua_State * script_L;
static LandArray * _per_thread;
struct Script {
    char * buffer;
    bool read;
};
static jmp_buf jump_environment;
static LandArray * errors;
static void* alloc(void * ud, void * ptr, size_t osize, size_t nsize);
static int panicf(lua_State * l);
static int fun_load_lua(lua_State * l);
static const char* reader(lua_State * L, void * data, size_t * size);
static void* alloc(void * ud, void * ptr, size_t osize, size_t nsize) {
    if (nsize == 0) {
        if (ptr) {
            land_free(ptr);
        }
        return NULL;
    }
    else {
        if (ptr) {
            return land_realloc(ptr, nsize);
        }
        else {
            return land_malloc(nsize);
        }
    }
}
static int panicf(lua_State * l) {
    print("lua panic: %s", lua_tostring(l, - 1));
    longjmp(jump_environment, 1);
    return 0;
}
static int fun_load_lua(lua_State * l) {
    char const * name = lua_tostring(l, 1);
    if (DebugLogLua) {
        print("fun_load_lua %s", name);
    }
    char * path = land_strdup(name);
    land_replace_all(& path, ".", "/");
    land_concatenate(& path, ".lua");
    script_load(l, path);
    land_free(path);
    return 1;
}
lua_State* script_init(void) {
    if (! script_L) {
        script_L = script_init_thread(0);
    }
    return script_L;
}
void script_init_new(PerThread * pt) {
    pt->l = lua_newstate(alloc, NULL);
    pt->registered_functions = land_hash_new();
    pt->loaded_scripts_hash = land_hash_new();
    pt->loaded_scripts_order = land_array_new();
    lua_getstack(pt->l, 0, & pt->debug);
    luaL_openlibs(pt->l);
    lua_atpanic(pt->l, panicf);
    common_add_cleanup(script_cleanup);
    helper_add_fnc(pt->l, "load_lua", fun_load_lua);
    script_load(pt->l, "lua/common.lua");
    script_execute(pt->l, "lua/common.lua");
}
lua_State* script_init_thread(int tid) {
    if (! _per_thread) {
        _per_thread = land_array_new();
    }
    int n = land_array_count(_per_thread);
    for (int i = n; i < tid + 1; i += 1) {
        land_array_add(_per_thread, NULL);
    }
    PerThread * pt = land_array_get(_per_thread, tid);
    if (pt == NULL) {
        land_alloc(pt);
        land_array_replace_nth(_per_thread, tid, pt);
        print("Initializing Lua state for thread %d", tid);
        script_init_new(pt);
    }
    return pt->l;
}
lua_Debug* script_debug_info(lua_State * l) {
    PerThread * pt = script_get_per_thread(l);
    return & pt->debug;
}
PerThread* script_get_per_thread(lua_State * l) {
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(_per_thread);
        for (PerThread * pt = LandArrayIterator_item(_per_thread, &__iter0__); LandArrayIterator_next(_per_thread, &__iter0__); pt = LandArrayIterator_item(_per_thread, &__iter0__)) {
            if (! pt) {
                continue;
            }
            if (pt->l == l) {
                return pt;
            }
        }
    }
    return NULL;
}
void script_cleanup(void) {
    if (! _per_thread) {
        return ;
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(_per_thread);
        for (PerThread * pt = LandArrayIterator_item(_per_thread, &__iter0__); LandArrayIterator_next(_per_thread, &__iter0__); pt = LandArrayIterator_item(_per_thread, &__iter0__)) {
            if (! pt) {
                continue;
            }
            lua_close(pt->l);
            {
                LandHashIterator __iter1__ = LandHashIterator_first(pt->loaded_scripts_hash);
                for (Script * s = LandHashIterator_item(pt->loaded_scripts_hash, &__iter1__); LandHashIterator_next(pt->loaded_scripts_hash, &__iter1__); s = LandHashIterator_item(pt->loaded_scripts_hash, &__iter1__)) {
                    land_free(s->buffer);
                    land_free(s);
                }
            }
            land_hash_destroy(pt->loaded_scripts_hash);
            land_array_destroy_with_strings(pt->loaded_scripts_order);
        }
    }
    land_array_destroy(_per_thread);
    _per_thread = NULL;
    script_L = NULL;
}
static const char* reader(lua_State * L, void * data, size_t * size) {
    Script * s = data;
    if (s->read) {
        return NULL;
    }
    s->read = 1;
    * size = strlen(s->buffer);
    return s->buffer;
}
void script_reload(lua_State * l, char const * name) {
    PerThread * pt = script_get_per_thread(l);
    Script * s = land_hash_get(pt->loaded_scripts_hash, name);
    if (s) {
        land_free(s->buffer);
        s->buffer = land_strdup("");
        script_load(l, name);
        script_execute(l, name);
    }
}
void script_error(char const * format, ...) {
    va_list args;
    va_start(args, format);
    int n = vsnprintf(NULL, 0, format, args);
    va_end(args);
    if (n < 0) {
        n = 1023;
    }
    char * e = land_malloc(n + 1);
    va_start(args, format);
    vsnprintf(e, n + 1, format, args);
    va_end(args);
    if (! errors) {
        errors = land_array_new();
    }
    land_array_add(errors, e);
}
LandArray* script_get_errors(void) {
    return errors;
}
void script_clear_errors(void) {
    land_array_clear(errors);
}
// [] -> [ function ]
bool script_load(lua_State * L, char const * name) {
    char * new_buffer = land_read_text(name);
    if (! new_buffer) {
        print("Could not load %s.", name);
        return 0;
    }
    bool ret = script_with_code(L, name, new_buffer);
    land_free(new_buffer);
    return ret;
}
void script_unload(lua_State * L, char const * name) {
    PerThread * pt = script_get_per_thread(L);
    Script * s = land_hash_get(pt->loaded_scripts_hash, name);
    if (! s) {
        return ;
    }
    land_free(s->buffer);
    land_hash_remove(pt->loaded_scripts_hash, name);
    int r = land_array_find_string(pt->loaded_scripts_order, name);
    land_free(land_array_get(pt->loaded_scripts_order, r));
    land_array_shift_remove(pt->loaded_scripts_order, r);
}
void script_register_same_functions_as(lua_State * l, lua_State * las) {
    PerThread * pt = script_get_per_thread(l);
    PerThread * laspt = script_get_per_thread(las);
    {
        LandHashKeysIterator __iter0__ = LandHashKeysIterator_first(laspt->registered_functions);
        for (char * name = LandHashKeysIterator_item(laspt->registered_functions, &__iter0__); LandHashKeysIterator_next(laspt->registered_functions, &__iter0__); name = LandHashKeysIterator_item(laspt->registered_functions, &__iter0__)) {
            if (! land_hash_has(pt->registered_functions, name)) {
                void * fun = land_hash_get(laspt->registered_functions, name);
                helper_add_fnc(l, name, fun);
            }
        }
    }
}
void script_load_same_files_as(lua_State * l, lua_State * las) {
    PerThread * pt = script_get_per_thread(l);
    PerThread * laspt = script_get_per_thread(las);
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(laspt->loaded_scripts_order);
        for (char * name = LandArrayIterator_item(laspt->loaded_scripts_order, &__iter0__); LandArrayIterator_next(laspt->loaded_scripts_order, &__iter0__); name = LandArrayIterator_item(laspt->loaded_scripts_order, &__iter0__)) {
            if (! land_hash_has(pt->loaded_scripts_hash, name)) {
                script_load(l, name);
                script_execute(l, name);
            }
        }
    }
}
// [] -> [ function ]
// Return value indicates whether the contents changed since the last
// time. It does not indicate failure or success!
bool script_with_code(lua_State * L, str name, str code) {
    PerThread * pt = script_get_per_thread(L);
    bool reload = 0;
    Script * s = land_hash_get(pt->loaded_scripts_hash, name);
    if (s) {
        if (strcmp(s->buffer, code)) {
            land_free(s->buffer);
            s->buffer = land_strdup(code);
            reload = 1;
        }
    }
    else {
        s = land_calloc(sizeof (* s));
        land_array_add(pt->loaded_scripts_order, land_strdup(name));
        land_hash_insert(pt->loaded_scripts_hash, name, s);
        s->buffer = land_strdup(code);
        reload = 1;
    }
    lua_getglobal(L, "scripts");
    if (lua_isnil(L, - 1)) {
        // [ nil ]
        lua_pop(L, 1);
        lua_createtable(L, 0, 0);
        lua_pushvalue(L, 1);
        // [ scripts scripts ]
        lua_setglobal(L, "scripts");
    }
    // [ scripts ]
    if (reload) {
        if (DebugLogLua) {
            print("loading script %s", name);
        }
        s->read = 0;
        str name2 = strrchr(name, '/');
        if (name2) {
            name2++;
        }
        else {
            name2 = name;
        }
        int e = lua_load(L, reader, s, name2, NULL);
        if (e) {
            char const * mess = lua_tostring(L, - 1);
            print("Lua loading error: %s", mess);
            lua_pushnil(L);
            script_error("%s", mess);
            return 1;
        }
        // [ scripts func]
        lua_pushstring(L, name);
        // [ scripts func name ]
        lua_pushvalue(L, - 2);
        // [ scripts func name func ]
        lua_settable(L, - 4);
        // [ scripts func ]
        lua_remove(L, - 2);
        // [ func ]
        return 1;
    }
    lua_pushstring(L, name);
    // [ scripts name ]
    lua_gettable(L, - 2);
    // [ scripts func ]
    lua_remove(L, - 2);
    // [ func ]
    return 0;
}
// [ function ] -> [ return_values... ]
void script_execute(lua_State * L, str name) {
    if (setjmp(jump_environment) == 0) {
        script_execute_params(L, name, 0);
    }
}
// [function parameter1 parameter2 ...] -> [retval1 retval2 ...]
void script_execute_params(lua_State * L, str name, int params) {
    // [x1...xm f p1...pn] -> handler_pos = m + 1 + n - n = m + 1
    int handler_pos = lua_gettop(L) - params /* (positive absolute index) */;
    lua_getglobal(L, "debug") /* [f p1...pn debug] */;
    lua_getfield(L, - 1, "traceback") /* [f p1...pn debug traceback] */;
    lua_remove(L, - 2) /* [f p1...pn traceback] */;
    lua_insert(L, handler_pos) /* [eh f p1...pn] */;
    int e = lua_pcall(L, params, LUA_MULTRET, handler_pos) /* [r1...rn] */;
    if (e) {
        char const * mess = lua_tostring(L, - 1);
        str errtype = "unknown";
        if (e == LUA_ERRRUN) {
            errtype = "runtime";
        }
        if (e == LUA_ERRMEM) {
            errtype = "memory";
        }
        if (e == LUA_ERRERR) {
            errtype = "handler";
        }
        print("Lua runtime error: %s: %s: %s", name, errtype, mess);
        land_log_message("script_execute: %s: %s: %s\n", name, errtype, mess);
        script_error("%s", mess);
    }
    lua_remove(L, handler_pos);
}
void script_debug(lua_State * L, int ind) {
    lua_pushnil(L);
    while (lua_next(L, ind)) {
        int kt = lua_type(L, - 2);
        int vt = lua_type(L, - 1);
        if (kt == LUA_TSTRING) {
            printf("%s", lua_tostring(L, - 2));
        }
        else {
            printf("?");
        }
        printf("=");
        if (vt == LUA_TSTRING) {
            printf("%s", lua_tostring(L, - 1));
        }
        else if (vt == LUA_TNUMBER) {
            printf("%g", lua_tonumber(L, - 1));
        }
        else {
            printf("?");
        }
        printf("\n");
        lua_pop(L, 1);
    }
}
bool script_debug_repeat(lua_State * L, LandArray * path, int repeat, char * key, char * val) {
    /* If repeat is 0, start iterating. The iterator will be left on the lua stack.
     * If repeat is not 0, iterate the next item.
     * Returns False when there is nothing left to iterate over and the stack has
     * been cleared again.
     * If path is none, this iterates over all global variables.
     * If it has a single entry, it is the name of a global table whose members
     * are looped over.
     * Additional entries can be used to loop tables inside tables.
     */
    if (repeat == 0) {
        int i = 0, n = land_array_count(path);
        lua_pushglobaltable(L);
        while (i < n) {
            for (int j = 0; j < i; j += 1) {
                printf(" ");
            }
            char * s = land_array_get_nth(path, i);
            printf("%s\n", s);
            lua_getfield(L, - 1, s);
            lua_remove(L, - 2);
            i++;
        }
        if (! lua_istable(L, - 1)) {
            lua_pop(L, 1);
            return 0;
        }
        lua_pushnil(L);
    }
    bool r = lua_next(L, - 2);
    if (! r) {
        lua_pop(L, 1);
        return 0;
    }
    int kt = lua_type(L, - 2);
    int vt = lua_type(L, - 1);
    if (kt == LUA_TSTRING) {
        sprintf(key, "%s", lua_tostring(L, - 2));
    }
    else if (kt == LUA_TNUMBER) {
        sprintf(key, "%g", lua_tonumber(L, - 2));
    }
    else {
        sprintf(key, "?");
    }
    if (vt == LUA_TSTRING) {
        sprintf(val, "%s", lua_tostring(L, - 1));
    }
    else if (vt == LUA_TNUMBER) {
        sprintf(val, "%g", lua_tonumber(L, - 1));
    }
    else if (vt == LUA_TBOOLEAN) {
        sprintf(val, "%d", lua_toboolean(L, - 1));
    }
    else if (vt == LUA_TTABLE) {
        sprintf(val, "{table}");
    }
    else if (vt == LUA_TFUNCTION) {
        sprintf(val, "(function)");
    }
    else if (vt == LUA_TNIL) {
        sprintf(val, "(none)");
    }
    else if (vt == LUA_TUSERDATA) {
        sprintf(val, "(user)");
    }
    else if (vt == LUA_TLIGHTUSERDATA) {
        sprintf(val, "(pointer)");
    }
    else {
        sprintf(val, "(type %d)", vt);
    }
    lua_pop(L, 1);
    return 1;
}
void script_startup(void) {
    lua_State * l = script_init();
    script_load(l, "lua/startup.lua");
    script_execute(l, "startup");
    script_unload(l, "lua/startup.lua") /* so other threads won't load it */;
}
