#include "world.h"
    /* = Coordinates
     * The world coordinate system has the world in the XY plane with Z going up.
     * 1 unit is as big as the size of 1 tile in the landscape grid.
     * The local coordinate system of actors also has Z going up. Y points to their
     * back, X to their left (or to the right when viewing them from the front).
     */
static LandColor _color(Landscape * self, float h);
static bool _find_between(LandVector a, LandVector b, LandVector * r, bool recurse);
void pos_to_hex(float px, float py, int * hx, int * hy) {
    * hx = floor(px / 4.0);
    if (* hx & 1) {
        py += 2;
    }
    * hy = floor(py / 4.0);
}
void hex_to_int(int hx, int hy, int * ix, int * iy) {
    * ix = hx * 4 + 2;
    * iy = hy * 4 + 2;
    if (hx & 1) {
        * iy -= 2;
    }
}
void add_dir(int * x, int * y, int d) {
    /* dir:
     * __/1 \__
     * /0 \__/2 \
     * \__/  \__/
     * /5 \__/3 \
     * \__/4 \__/
     * \__/
     * x: -1  0  1  2
     * y:     __    __
     * __/ 1\__/ 1\
     * / 1\__/ 1\__/
     * \__/ 0\__/ 0\
     * /0 \__/ 0\__/
     * \__/-1\__/-1\
     * \__/  \__/
     */
    int dx [] = {- 1, 0, 1, 1, 0, - 1};
    int dy [2] [6] = {{1, 1, 1, 0, - 1, 0}, {0, 1, 0, - 1, - 1, - 1}};
    * y += dy [* x & 1] [d];
    * x += dx [d];
}
int get_hex_dir(float dx, float dy) {
    float a = atan2(- dy, dx) - pi * 2 * 7 / 12;
    a = round(a * 3 / pi);
    int i = a;
    i %= 6;
    if (i < 0) {
        return i + 6;
    }
    return i;
}
float get_top_z(int x, int y) {
    Application * app = app_global();
    Landscape * ls = app->world->landscape;
    float z = landscape_get_height_interpolated(ls, x, y);
    int ix, iy;
    pos_to_hex(x, y, & ix, & iy);
    int hz = find_highest_z(ix, iy);
    if (hz * 4 > z) {
        return hz * 4;
    }
    return z;
}
Land4x4Matrix world_get_projection_matrix(void) {
    float w = land_display_width();
    float h = land_display_height();
    return land_4x4_matrix_orthographic(- w / 2, - h / 2, 1000, w / 2, h / 2, - 1000);
}
int world_width(void) {
    Application * app = app_global();
    return app->world->landscape->w;
}
void world_render(World * self, Camera * cam) {
    Application * app = app_global();
    Landscape * ls = self->landscape;
    if (! app->debug_no_landscape) {
        int s = ls->max_visible_chunks_s;
        LandFloat z = pow(2, cam->zoom);
        s = 4 + (s - 4) * 4 / z;
        ls->visible_chunks_w = s;
        ls->visible_chunks_h = s;
        landscape_render(ls, cam);
    }
    bricks_draw(cam, 0, 0);
    float xo = app->cam.p.x < ls->w / 2 ? - ls->w : ls->w;
    float yo = app->cam.p.y < ls->w / 2 ? - ls->w : ls->w;
    bricks_draw(cam, xo, 0);
    bricks_draw(cam, xo, yo);
    bricks_draw(cam, 0, yo);
}
World* world_new_with_landscape(Landscape * ls) {
    World * world;
    land_alloc(world);
    app_global ()->world = world;
    world->landscape = ls;
    return world;
}
static LandColor _color(Landscape * self, float h) {
    if (h < 2) {
        float x = h / 2;
        return land_color_rgba(0.9 * x, 0.9 * x, 0.8 * x, x);
    }
    return land_color_rgba(0.9, 0.9, 0.8, 1);
}
void create_map(Landscape * self, int seed) {
    Application * app = app_global();
    int w = self->w;
    self->heightmap = land_calloc(w * w * sizeof (* self->heightmap));
    // limit is -1..+1, so we divide amplitude by half the maximum value
    LandNoise * noise = land_noise_new(LandNoisePerlin, seed);
    land_noise_set_lerp(noise, LandPerlinLerpCosine);
    land_noise_set_size(noise, w, w);
    land_noise_set_levels(noise, 4);
    land_noise_set_first_level(noise, 3);
    land_noise_set_power_modifier(noise, 0);
    land_noise_set_amplitude(noise, 1.2);
    land_noise_z_transform(noise, 1, - 1.2);
    land_noise_z_ease(noise, 0);
    LandNoise * blur = noise /* land_noise_new(LandNoiseNone, 0) */;
    //land_noise_set_random(blur, noise.seed)
    //land_noise_set_blur(blur, noise, self.blur)
    land_noise_prepare(blur);
    int highest = 0;
    int hx = 0, hy = 0;
    for (int y = 0; y < w; y += 1) {
        for (int x = 0; x < w; x += 1) {
            Tile * tile = self->heightmap + y * w + x;
            float v = land_noise_at(blur, x, y);
            v = (1 + v) / 2;
            if (v < 0) {
                v = 0;
            }
            if (v > 1) {
                v = 1;
            }
            tile->height = self->max_height * v;
            if (tile->height > highest) {
                highest = tile->height;
                hx = x;
                hy = y;
            }
        }
    }
    app->start_x = hx;
    app->start_y = hy;
    int want_distance = 100 + app->level * 50;
    int best_score = 0;
    for (int y = 0; y < w; y += 1) {
        for (int x = 0; x < w; x += 1) {
            Tile * tile = self->heightmap + y * w + x;
            float dx = rwrap(x - app->start_x, w);
            float dy = rwrap(y - app->start_y, w);
            float d = sqrt(dx * dx + dy * dy);
            float dd = fabs(d - want_distance);
            int score = tile->height - dd;
            if (tile->height < 1) {
                score = 0;
            }
            if (score > best_score) {
                best_score = score;
                app->goal_x = x;
                app->goal_y = y;
            }
        }
    }
}
World* world_new(void) {
    Application * app = app_global();
    Landscape * l = landscape_create_base();
    l->max_visible_chunks_s = 11;
    l->visible_chunks_h = l->max_visible_chunks_s;
    l->visible_chunks_w = l->max_visible_chunks_s;
    l->chunks_count = l->visible_chunks_h * l->visible_chunks_w;
    l->chunk_s = 128;
    l->w = 1 << 9;
    l->max_height = 20;
    l->snow_top = 11;
    l->water_levels = 0 /* -128 */;
    l->plateau_levels = 0;
    l->want_rivers = 0;
    l->color_for_height = _color;
    l->blur = 2;
    l->fog_color = land_color_rgba(0, 0, 0, 1);
    print("creating map %lu", app->seed);
    create_map(l, app->seed);
    for (int y = 0; y < l->w; y += 1) {
        for (int x = 0; x < l->w; x += 1) {
            Tile * tile = l->heightmap + y * l->w + x;
            int rx = x;
            int ry = y;
            if (rx > l->w / 2) {
                rx -= l->w;
            }
            if (ry > l->w / 2) {
                ry -= l->w;
            }
            if (rx * rx + ry * ry > 240 * 240) {
                float v = (sqrt(rx * rx + ry * ry) - 240) / 16.0;
                float h = tile->height * (1 - v);
                if (h < 0) {
                    h = 0;
                }
            }
        }
    }
    //tile.height = h
    landscape_calculate_light(l);
    landscape_create_colors(l);
    landscape_make_circle_visible(l, 0, 0, 400, 400);
    World * w = world_new_with_landscape(l);
    w->s = (int)(l->w / 4);
    if (w->bricks_map) {
        land_free(w->bricks_map);
    }
    w->bricks_map = land_calloc(w->s * w->s * 32);
    if (w->bricks) {
        clear_bricks();
    }
    w->bricks = land_array_new();
    make_wall_hex(w);
    make_bridge_hex(w);
    //for int i in range(21):
    //int p = (-10 + i) * 4
    //add_brick(Wall, p + 2, -10 * 4 + 2)
    //add_brick(Wall, p + 2, 10 * 4 + 2)
    //add_brick(Wall, -10 * 4 + 2, p + 2)
    //add_brick(Wall, 10 * 4 + 2, p + 2)
    units_clear();
    Unit * u = unit_new(UnitMothership, app->start_x, app->start_y);
    camera_change_freely(& u->cam, 0, 0, land_rnd(0, pi * 2));
    u->cam.p.z = 30;
    u->start = u->cam.p;
    level();
    return w;
}
void level(void) {
    Application * app = app_global();
    app->elephants = 10 + 10 * app->level;
}
void world_destroy(World * self) {
    landscape_destroy(self->landscape);
    land_free(self);
}
static bool _find_between(LandVector a, LandVector b, LandVector * r, bool recurse) {
    LandVector last_above = a;
    Application * app = app_global();
    World * world = app->world;
    LandVector d = land_vector_sub(b, a);
    for (int i = 0; i < 10; i += 1) {
        LandVector v = land_vector_add(a, land_vector_mul(d, i / 9.0));
        LandFloat h = landscape_get_height_interpolated(world->landscape, v.x, v.y);
        int x, y;
        pos_to_hex(v.x, v.y, & x, & y);
        //LandFloat h = get_top_z(x, y)
        if (v.z > h) {
            last_above = v;
        }
        else {
            if (recurse) {
                return _find_between(last_above, v, r, 0);
            }
            else {
                * r = v;
                return 1;
            }
        }
    }
    return 0;
}
LandVector world_get_position_under_z(Camera * c, float mx, float my) {
    Application * app = app_global();
    World * world = app->world;
    float w = land_display_width();
    float h = land_display_height();
    LandVector t;
    t.x = mx - w / 2;
    t.y = - my + h / 2;
    t.z = 0;
    LandFloat s = pow(2, c->zoom);
    t.x /= s;
    t.y /= s;
    t.z /= s;
    t = land_vector_backtransform(t, c->p, c->x, c->y, c->z);
    // t is a position on our view plane now.
    //
    //      /
    //  -  /
    //    o
    //   /  -
    //  /_____O_____
    //
    LandFloat zeroheight = (0 - t.z) / c->z.z;
    LandFloat maxheight = (world->landscape->max_height - t.z) / c->z.z;
    _find_between(land_vector_add(t, land_vector_mul(c->z, maxheight)), land_vector_add(t, land_vector_mul(c->z, zeroheight)), & t, 1);
    return t;
}
